import io
import logging
import os
import re
import shutil
import sys
from argparse import ArgumentParser

import alexfitter
import pandas as pd
from numpy import arange
from contextlib import redirect_stdout


def get_from_user(target, key):
    value = None
    if isinstance(target, dict) and key in target:
        value = target[key]
    return value


class TtvFitter:
    def __init__(self, only_initial, ttvs_error, fit_dir, ttvs_dir):
        self.only_initial = only_initial
        self.fit_dir = fit_dir
        self.ttvs_dir = ttvs_dir
        self.ttvs_error = ttvs_error

    def fit(self):
        shutil.copy(self.fit_dir + "/lc.csv", self.ttvs_dir + "/lc.csv")
        shutil.copy(self.fit_dir + "/params.csv", self.ttvs_dir + "/params.csv")
        shutil.copy(self.fit_dir + "/params_star.csv", self.ttvs_dir + "/params_star.csv")
        shutil.copy(self.fit_dir + "/settings.csv", self.ttvs_dir + "/settings.csv")
        self._tune_settings()
        self._fix_time_params()
        self._prepare_ttv_params()
        alexfitter.show_initial_guess(ttvs_dir)
        if not args.only_initial:
            logging.info("Running dynamic nested sampling")
            alexfitter.ns_fit(ttvs_dir)
            alexfitter.ns_output(ttvs_dir)

    def _tune_settings(self):
        logging.info("Enabling TTVs fitting setting")
        with open(ttvs_dir + "/settings.csv", 'r+') as f:
            text = f.read()
            text = re.sub('fit_ttvs,False', 'fit_ttvs,True', text)
            f.seek(0)
            f.write(text)
            f.truncate()

    def _fix_time_params(self):
        logging.info("Fixing epochs and periods")
        params_df = pd.read_csv(ttvs_dir + "/params.csv")
        ns_table_df = pd.read_csv(object_dir + "/results/ns_table.csv")
        candidates_count = len(ns_table_df[ns_table_df["#name"].str.contains("_period")])
        for i in arange(0, candidates_count):
            period_row = ns_table_df[ns_table_df["#name"].str.contains("_period")].iloc[i]
            period_row_name = period_row["#name"]
            period = period_row["median"]
            epoch_row = ns_table_df[ns_table_df["#name"].str.contains("_epoch")].iloc[i]
            epoch_row_name = epoch_row["#name"]
            epoch = epoch_row["median"]
            params_df.loc[params_df["#name"] == period_row_name, "value"] = period
            params_df.loc[params_df["#name"] == period_row_name, "fit"] = 0
            params_df.loc[params_df["#name"] == epoch_row_name, "value"] = epoch
            params_df.loc[params_df["#name"] == epoch_row_name, "fit"] = 0
        params_df.to_csv(ttvs_dir + "/params.csv", index=False)

    def _prepare_ttv_params(self):
        logging.info("Preparing TTVs params")
        alexfitter.prepare_ttv_fit(self.ttvs_dir)
        with open(self.ttvs_dir + '/ttv_preparation/ttv_initial_guess_params.csv') as f:
            ttvs_params = f.readlines()
        ttvs_error_str = str(self.ttvs_error)
        ttvs_error_str = ",uniform -" + ttvs_error_str + " " + ttvs_error_str + ","
        ttvs_initial_str = ",0.0,1,uniform"
        ttvs_params = [re.sub(r",uniform -?\d+(\.\d+)? -?\d+(\.\d+)?,", ttvs_error_str, ttvs_param) for ttvs_param in ttvs_params]
        ttvs_params = [re.sub(r",-?\d+(\.\d+)?,1,uniform", ttvs_initial_str, ttvs_param) for ttvs_param in ttvs_params]
        with open(self.ttvs_dir + "/params.csv", "a") as params_file:
            params_file.write("\n" + "".join(ttvs_params))


if __name__ == '__main__':
    ap = ArgumentParser(description='Calculation of Time Transits Variations')
    ap.add_argument('--object_dir', help="If the object directory is not your current one you need to provide the "
                                         "ABSOLUTE path", required=False)
    ap.add_argument('--cpus', type=int, default=4, help="The number of CPU cores to be used.", required=False)
    ap.add_argument('--only_initial', dest='only_initial', action='store_true',
                    help="Whether to only run an initial guess of the ttvs")
    ap.add_argument('--ttvs_error', type=float, default=0.07, required=False,
                    help="Error window for each single transit fit")
    args = ap.parse_args()
    object_dir = os.getcwd() if args.object_dir is None else args.object_dir
    index = 0
    ttvs_dir = object_dir + "/ttvs_" + str(index)
    while os.path.exists(ttvs_dir) or os.path.isdir(ttvs_dir):
        ttvs_dir = object_dir + "/ttvs_" + str(index)
        index = index + 1
    os.mkdir(ttvs_dir)
    file_dir = ttvs_dir + "/ttvs.log"
    if os.path.exists(file_dir):
        os.remove(file_dir)
    if not isinstance(logging.root, logging.RootLogger):
        logging.root = logging.RootLogger(logging.INFO)
    formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger()
    while len(logger.handlers) > 0:
        logger.handlers.pop()
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    handler = logging.FileHandler(file_dir)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logging.info("Starting TTVs computation")
    TtvFitter(args.only_initial, args.ttvs_error, object_dir, ttvs_dir).fit()



import math
from sherlockpipe.search_zones.SearchZone import SearchZone
from lcbuilder.star.HabitabilityCalculator import HabitabilityCalculator
from lcbuilder.star.starinfo import StarInfo


class OptimisticHabitableSearchZone(SearchZone):
    """
    Implementation for calculating minimum and maximum search periods for the habitable zone of an input star.
    """
    def __init__(self):
        super().__init__()

    def calculate_period_range(self, star_info: StarInfo):
        """
        Calculates the minimum and maximum periods for the Optimistic Habitable Zone (from kopparapu et al 2013)
        given the star_info.

        :param star_info: the star where the range should be calculated
        :return: a tuple of minimum_period and maximum_period
        """
        habitability_calc = HabitabilityCalculator()
        hz_periods = habitability_calc.calculate_hz_periods(star_info.teff, star_info.lum, star_info.mass)
        if hz_periods is None:
            return None
        return hz_periods[0], hz_periods[3]



import os
import shutil
import types
import unittest
import pkg_resources
from astropy.time import Time

from sherlockpipe.observation_plan.run import run_plan
from sherlockpipe.search.run import run_search
from sherlockpipe.system_stability.run import run_stability
from sherlockpipe.validation.run import run_validate
from sherlockpipe.vetting.run import run_vet


class TestsEntrypoints(unittest.TestCase):
    def test_validation(self):
        object_dir = TestsEntrypoints.get_path('test_endpoints_data/')
        args = types.SimpleNamespace()
        args.object_dir = object_dir
        args.candidate = 1
        args.cpus = 1
        args.contrast_curve = None
        args.bins = 50
        args.scenarios = 2
        args.sigma_mode = 'flux_err'
        validation_dir = object_dir + '/validation_0'
        try:
            run_validate(args)
            self.assertEquals(6, len(os.listdir(validation_dir + '/triceratops')))
        finally:
            shutil.rmtree(validation_dir, ignore_errors=True)

    def test_vetting(self):
        object_dir = TestsEntrypoints.get_path('test_endpoints_data/')
        vetting_dir = object_dir + '/vetting_0'
        try:
            run_vet(object_dir, 1, None, cpus=4)
            self.assertEquals(7, len(os.listdir(vetting_dir)))
        finally:
            shutil.rmtree(vetting_dir, ignore_errors=True)

    def test_stability(self):
        object_dir = TestsEntrypoints.get_path('test_endpoints_data/')
        properties_dir = TestsEntrypoints.get_path("test_endpoints_data/stability.yaml")
        args = types.SimpleNamespace()
        args.object_dir = object_dir
        args.properties = properties_dir
        args.cpus = 4
        args.star_mass_bins = 1
        args.period_bins = 1
        args.free_params = None
        args.use_spock = False
        args.years = 500
        try:
            run_stability(args)
            self.assertEquals(2, len(os.listdir(object_dir + '/stability_0')))
        finally:
            shutil.rmtree(object_dir + '/stability_0', ignore_errors=True)

    def test_search(self):
        properties_dir = TestsEntrypoints.get_path("search.yaml")
        results_dir = TestsEntrypoints.get_path("/")
        search_dir = results_dir + "TIC305048087_[2]"
        try:
            run_search(properties_dir, False, results_dir, 4)
            self.assertEquals(20, len(os.listdir(search_dir)))
        finally:
            shutil.rmtree(search_dir, ignore_errors=True)

    @unittest.skip("Allesfitter is failing loading data: https://github.com/MNGuenther/allesfitter/issues/57")
    def test_plan(self):
        object_dir = TestsEntrypoints.get_path('test_endpoints_data/fit_1/')
        plan_dir = object_dir + 'plan/'
        args = types.SimpleNamespace()
        args.object_dir = object_dir
        args.cpus = 4
        args.observatories = object_dir + 'observatories.csv'
        args.since = Time('2022-10-01', scale='utc')
        args.error_sigma = 1
        args.time_unit = None
        args.tz = 0
        args.lat = None
        args.lon = None
        args.alt = None
        args.max_days = 30
        args.min_altitude = 25
        args.moon_min_dist = 20
        args.moon_max_dist = 40
        args.transit_fraction = 0.5
        args.no_error_alert = True
        args.baseline = 1
        try:
            run_plan(args)
            self.assertEquals(2, len(os.listdir(plan_dir)))
            with open(plan_dir + "observation_plan.csv", 'r') as fp:
                self.assertEquals(8, len(fp.readlines()))
        finally:
            shutil.rmtree(plan_dir, ignore_errors=True)


    @staticmethod
    def get_path(path):
        """
        Gets right path for tests environment
        :param path:
        :return: the real path of the test resource
        """
        return pkg_resources.resource_filename(__name__, path)


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod

from lcbuilder.star.starinfo import StarInfo


class SearchZone(ABC):
    """
    Abstract class to be implemented for calculating minimum and maximum search periods for an input star.
    """
    def __init__(self):
        pass

    @abstractmethod
    def calculate_period_range(self, star_info: StarInfo):
        """
        Calculates the minimum and maximum periods for the given star_info

        :param star_info: the star where the range should be calculated
        :return: a tuple of minimum_period and maximum_period
        """
        pass



from argparse import ArgumentParser

from sherlockpipe.validation.run import run_validate


def validation_args_parse(arguments=None):
    ap = ArgumentParser(description='Validation of Sherlock objects of interest')
    ap.add_argument('--object_dir', help="If the object directory is not your current one you need to provide the "
                                         "ABSOLUTE path", required=False)
    ap.add_argument('--candidate', type=int, default=None, help="The candidate signal to be used.", required=False)
    ap.add_argument('--properties', help="The YAML file to be used as input.", required=False)
    ap.add_argument('--cpus', type=int, default=None, help="The number of CPU cores to be used.", required=False)
    ap.add_argument('--bins', type=int, default=100, help="The number of bins to be used for the folded curve "
                                                          "validation.", required=False)
    ap.add_argument('--sigma_mode', type=str, default='flux_err', help="The way to calculate the sigma value for the "
                                                                       "validation. [flux_err|binning]", required=False)
    ap.add_argument('--scenarios', type=int, default=5, help="The number of scenarios to be used for the validation",
                    required=False)
    ap.add_argument('--contrast_curve', type=str, default=None, help="The contrast curve in csv format.",
                    required=False)
    return ap.parse_args(arguments)


if __name__ == '__main__':
    args = validation_args_parse()
    run_validate(args)



from argparse import ArgumentParser
from sherlockpipe.search.run import run_search


if __name__ == '__main__':
    ap = ArgumentParser(description='Searching for Hints of Exoplanets fRom Lightcurves Of spaCe-based seeKers')
    ap.add_argument('--properties', help="Additional properties to be loaded into Sherlock run ", required=True)
    ap.add_argument('--results_dir', help="Directory where results should be written to", required=False, default=None)
    ap.add_argument('--explore', dest='explore', action='store_true',
                    help="Whether to run using mcmc or ns. Default is ns.")
    args = ap.parse_args()
    run_search(args.properties, args.explore, args.results_dir)



import logging
import os
import sys
import numpy as np
import pandas as pd
from lcbuilder.star.HabitabilityCalculator import HabitabilityCalculator
from sherlockpipe.bayesian_fit.fitter import Fitter
from sherlockpipe.bayesian_fit.fitter2 import Fitter2
from sherlockpipe.loading import common


def run_fit(args):
    index = 0
    object_dir = os.getcwd() if args.object_dir is None else args.object_dir
    candidates_df = pd.read_csv(object_dir + "/candidates.csv")
    candidate_selections = []
    if args.candidate is not None:
        candidate_selections = str(args.candidate)
        candidate_selections = candidate_selections.split(",")
        candidate_selections = list(map(int, candidate_selections))
    fitting_dir = object_dir + "/fit_" + str(index)
    while os.path.exists(fitting_dir) or os.path.isdir(fitting_dir):
        fitting_dir = object_dir + "/fit_" + str(index)
        index = index + 1
    os.mkdir(fitting_dir)
    file_dir = fitting_dir + "/fit.log"
    if os.path.exists(file_dir):
        os.remove(file_dir)
    if not isinstance(logging.root, logging.RootLogger):
        logging.root = logging.RootLogger(logging.INFO)
    formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger()
    while len(logger.handlers) > 0:
        logger.handlers.pop()
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    handler = logging.FileHandler(file_dir)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    star_df = pd.read_csv(object_dir + "/params_star.csv")
    if args.candidate is None:
        user_properties = common.load_from_yaml(args.properties)
        selected_candidates_df = pd.DataFrame(
            columns=['id', 'period', 't0', 'duration', 'cpus', 'rp_rs', 'a', 'number', 'name',
                     'lc'])
        selected_candidates_df = selected_candidates_df.append(user_properties["planets"], ignore_index=True)
        user_star_df = pd.DataFrame(columns=['R_star', 'M_star'])
        if "star" in user_properties and user_properties["star"] is not None:
            user_star_df = user_star_df.append(user_properties["star"], ignore_index=True)
            if user_star_df.iloc[0]["R_star"] is not None:
                star_df.at[0, "R_star"] = user_star_df.iloc[0]["R_star"]
            if user_star_df.iloc[0]["M_star"] is not None:
                star_df.at[0, "M_star"] = user_star_df.iloc[0]["M_star"]
            if user_star_df.iloc[0]["ld_a"] is not None:
                star_df.at[0, "ld_a"] = user_star_df.iloc[0]["ld_a"]
            if user_star_df.iloc[0]["ld_b"] is not None:
                star_df.at[0, "ld_b"] = user_star_df.iloc[0]["ld_b"]
            if ("a" not in user_properties["planet"] or user_properties["planet"]["a"] is None) \
                    and star_df.iloc[0]["M_star"] is not None and not np.isnan(star_df.iloc[0]["M_star"]):
                selected_candidates_df.at[0, "a"] = HabitabilityCalculator() \
                    .calculate_semi_major_axis(user_properties["planet"]["period"],
                                               user_properties["star"]["M_star"])
            elif ("a" not in user_properties["planet"] or user_properties["planet"]["a"] is None) \
                    and (star_df.iloc[0]["M_star"] is None or np.isnan(star_df.iloc[0]["M_star"])):
                raise ValueError("Cannot guess semi-major axis without star mass.")
        selected_candidates_df['number'] = user_properties["number"]
        selected_candidates_df['curve'] = user_properties["curve"]
        if selected_candidates_df.iloc[0]["a"] is None or np.isnan(selected_candidates_df.iloc[0]["a"]):
            raise ValueError("Semi-major axis is neither provided nor inferred.")
        if selected_candidates_df.iloc[0]["name"] is None:
            raise ValueError("You need to provide a name for your candidate.")
        if selected_candidates_df.iloc[0]["lc"] is None:
            raise ValueError("You need to provide a light curve relative path for your candidate.")
    else:
        selected_candidates_df = pd.read_csv(object_dir + "/candidates.csv")
        selected_candidates_df = selected_candidates_df.rename(columns={'Object Id': 'TICID'})
        selected_candidates_df["number"] = ""
        selected_candidates_df["name"] = ""
        for candidate_selection in candidate_selections:
            selected_candidates_df.loc[candidate_selection - 1, ['number']] = candidate_selection
            selected_candidates_df.loc[candidate_selection - 1, ['name']] = 'SOI_' + \
                                                                      str(selected_candidates_df['number'][
                                                                              candidate_selection - 1])
        selected_candidates_df = selected_candidates_df.iloc[
            [candidate_selection - 1 for candidate_selection in candidate_selections]]
        logging.info("Selected signal numbers " + str(candidate_selections))
    if args.cpus is None:
        cpus = os.cpu_count() - 1
    else:
        cpus = args.cpus
    fitter = Fitter(object_dir, fitting_dir, args.only_initial, len(selected_candidates_df) == 1, candidates_df,
                    args.mcmc, args.detrend)
    fitter.fit(selected_candidates_df, star_df, cpus, fitting_dir, args.tolerance, args.fit_orbit)

def run_fit2(args):
    index = 0
    object_dir = os.getcwd() if args.object_dir is None else args.object_dir
    candidates_df = pd.read_csv(object_dir + "/candidates.csv")
    candidate_selections = []
    if args.candidate is not None:
        candidate_selections = str(args.candidate)
        candidate_selections = candidate_selections.split(",")
        candidate_selections = list(map(int, candidate_selections))
    fitting_dir = object_dir + "/fit_" + str(index)
    while os.path.exists(fitting_dir) or os.path.isdir(fitting_dir):
        fitting_dir = object_dir + "/fit_" + str(index)
        index = index + 1
    os.mkdir(fitting_dir)
    file_dir = fitting_dir + "/fit.log"
    if os.path.exists(file_dir):
        os.remove(file_dir)
    if not isinstance(logging.root, logging.RootLogger):
        logging.root = logging.RootLogger(logging.INFO)
    formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger()
    while len(logger.handlers) > 0:
        logger.handlers.pop()
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    handler = logging.FileHandler(file_dir)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    star_df = pd.read_csv(object_dir + "/params_star.csv")
    if args.candidate is None:
        user_properties = common.load_from_yaml(args.properties)
        selected_candidates_df = pd.DataFrame(
            columns=['id', 'period', 't0', 'duration', 'cpus', 'rp_rs', 'a', 'number', 'name',
                     'lc'])
        selected_candidates_df = selected_candidates_df.append(user_properties["planets"], ignore_index=True)
        user_star_df = pd.DataFrame(columns=['R_star', 'M_star'])
        if "star" in user_properties and user_properties["star"] is not None:
            user_star_df = user_star_df.append(user_properties["star"], ignore_index=True)
            if user_star_df.iloc[0]["R_star"] is not None:
                star_df.at[0, "R_star"] = user_star_df.iloc[0]["R_star"]
            if user_star_df.iloc[0]["M_star"] is not None:
                star_df.at[0, "M_star"] = user_star_df.iloc[0]["M_star"]
            if user_star_df.iloc[0]["ld_a"] is not None:
                star_df.at[0, "ld_a"] = user_star_df.iloc[0]["ld_a"]
            if user_star_df.iloc[0]["ld_b"] is not None:
                star_df.at[0, "ld_b"] = user_star_df.iloc[0]["ld_b"]
            if ("a" not in user_properties["planet"] or user_properties["planet"]["a"] is None) \
                    and star_df.iloc[0]["M_star"] is not None and not np.isnan(star_df.iloc[0]["M_star"]):
                selected_candidates_df.at[0, "a"] = HabitabilityCalculator() \
                    .calculate_semi_major_axis(user_properties["planet"]["period"],
                                               user_properties["star"]["M_star"])
            elif ("a" not in user_properties["planet"] or user_properties["planet"]["a"] is None) \
                    and (star_df.iloc[0]["M_star"] is None or np.isnan(star_df.iloc[0]["M_star"])):
                raise ValueError("Cannot guess semi-major axis without star mass.")
        selected_candidates_df['number'] = user_properties["number"]
        selected_candidates_df['curve'] = user_properties["curve"]
        if selected_candidates_df.iloc[0]["a"] is None or np.isnan(selected_candidates_df.iloc[0]["a"]):
            raise ValueError("Semi-major axis is neither provided nor inferred.")
        if selected_candidates_df.iloc[0]["name"] is None:
            raise ValueError("You need to provide a name for your candidate.")
        if selected_candidates_df.iloc[0]["lc"] is None:
            raise ValueError("You need to provide a light curve relative path for your candidate.")
    else:
        selected_candidates_df = pd.read_csv(object_dir + "/candidates.csv")
        selected_candidates_df = selected_candidates_df.rename(columns={'Object Id': 'TICID'})
        selected_candidates_df["number"] = ""
        selected_candidates_df["name"] = ""
        for candidate_selection in candidate_selections:
            selected_candidates_df.loc[candidate_selection - 1, ['number']] = candidate_selection
            selected_candidates_df.loc[candidate_selection - 1, ['name']] = 'SOI_' + \
                                                                      str(selected_candidates_df['number'][
                                                                              candidate_selection - 1])
        selected_candidates_df = selected_candidates_df.iloc[
            [candidate_selection - 1 for candidate_selection in candidate_selections]]
        logging.info("Selected signal numbers " + str(candidate_selections))
    if args.cpus is None:
        cpus = os.cpu_count() - 1
    else:
        cpus = args.cpus
    if args.only_fit_candidate is None:
        only_fit_candidate_df = selected_candidates_df
    else:
        fit_candidate_numbers = [int(candidate_number) for candidate_number in str(args.only_fit_candidate).split(',')]
        only_fit_candidate_df = selected_candidates_df.loc[selected_candidates_df['number'].isin(fit_candidate_numbers)]
    fitter = Fitter2(object_dir, fitting_dir, args.only_initial, len(selected_candidates_df) == 1, candidates_df,
                    args.mcmc, args.detrend, rho_err_multi=args.rho_err_multi, sigma_err_multi=args.sigma_err_multi,
                     yerr_err_multi=args.yerr_err_multi)
    fitter.fit(selected_candidates_df, only_fit_candidate_df, star_df, cpus, fitting_dir, args.tolerance, args.fit_orbit)



import matplotlib.pyplot as plt
import numpy as np
from scipy import stats

from sherlockpipe.search.transitresult import TransitResult


def save_transit_plot(object_id: str, title: str, plot_dir: str, file: str, time, lc, transit_result: TransitResult,
                      cadence, run_no: int, plot_harmonics: bool = False):
    """
    Stores the search results plot with: 1) The entire curve with the transit model 2)The folded curve and the transit
    model 3) The power spectrum of the TLS search 4) Only if the flag is enabled, the TLS search harmonics power
    spectrum.

    :param str object_id: the target id
    :param str title: title for the plot
    :param str plot_dir: directory where the plot should be stored
    :param str file: the file name of the plot
    :param time: the time array of the light curve
    :param lc: the flux value of the light curve
    :param TransitResult transit_result: the TransitResult object containing the search results
    :param cadence: the cadence of the curve in days
    :param int run_no: the SHERLOCK run of the results
    :param bool plot_harmonics: whether the harmonics power spectrum should be plotted
    """
    # start the plotting
    rows = 3 if not plot_harmonics else 4
    figsize = 10 if not plot_harmonics else 14
    fig, axs = plt.subplots(nrows=rows, ncols=1, figsize=(10, figsize), constrained_layout=True)
    fig.suptitle(title)
    # 1-Plot all the transits
    in_transit = transit_result.in_transit
    tls_results = transit_result.results
    axs[0].scatter(time[in_transit], lc[in_transit], color='red', s=2, zorder=0)
    axs[0].scatter(time[~in_transit], lc[~in_transit], color='black', alpha=0.05, s=2, zorder=0)
    axs[0].plot(tls_results.model_lightcurve_time, tls_results.model_lightcurve_model, alpha=1, color='red', zorder=1)
    # plt.scatter(time_n, flux_new_n, color='orange', alpha=0.3, s=20, zorder=3)
    plt.xlim(time.min(), time.max())
    # plt.xlim(1362.0,1364.0)
    axs[0].set(xlabel='Time (days)', ylabel='Relative flux')
    # phase folded plus binning
    bins_per_transit = 8
    half_duration_phase = transit_result.duration / 2 / transit_result.period
    if np.isnan(transit_result.period) or np.isnan(transit_result.duration):
        bins = 200
        folded_plot_range = 0.05
    else:
        bins = transit_result.period / transit_result.duration * bins_per_transit
        folded_plot_range = half_duration_phase * 10
    binning_enabled = True
    axs[1].plot(tls_results.model_folded_phase, tls_results.model_folded_model, color='red')
    scatter_measurements_alpha = 0.05 if binning_enabled else 0.8
    axs[1].scatter(tls_results.folded_phase, tls_results.folded_y, color='black', s=10,
                alpha=scatter_measurements_alpha, zorder=2)
    lower_x_limit = 0.5 - folded_plot_range
    upper_x_limit = 0.5 + folded_plot_range
    axs[1].set_xlim(lower_x_limit, upper_x_limit)
    axs[1].set(xlabel='Phase', ylabel='Relative flux')
    folded_phase_zoom_mask = np.argwhere((tls_results.folded_phase > lower_x_limit) &
                                         (tls_results.folded_phase < upper_x_limit)).flatten()
    if isinstance(tls_results.folded_phase, (list, np.ndarray)):
        folded_phase = tls_results.folded_phase[folded_phase_zoom_mask]
        folded_y = tls_results.folded_y[folded_phase_zoom_mask]
        axs[1].set_ylim(np.min([np.min(folded_y), np.min(tls_results.model_folded_model)]),
                     np.max([np.max(folded_y), np.max(tls_results.model_folded_model)]))
        plt.ticklabel_format(useOffset=False)
        bins = 80
        if binning_enabled and tls_results.SDE != 0 and bins < len(folded_phase):
            bin_means, bin_edges, binnumber = stats.binned_statistic(folded_phase, folded_y, statistic='mean',
                                                                     bins=bins)
            bin_stds, _, _ = stats.binned_statistic(folded_phase, folded_y, statistic='std', bins=bins)
            bin_width = (bin_edges[1] - bin_edges[0])
            bin_centers = bin_edges[1:] - bin_width / 2
            bin_size = int(folded_plot_range * 2 / bins * transit_result.period * 24 * 60)
            bin_means_data_mask = np.isnan(bin_means)
            axs[1].errorbar(bin_centers[~bin_means_data_mask], bin_means[~bin_means_data_mask],
                         yerr=bin_stds[~bin_means_data_mask] / 2, xerr=bin_width / 2, marker='o', markersize=4,
                         color='darkorange', alpha=1, linestyle='none', label='Bin size: ' + str(bin_size) + "m")
            axs[1].legend(loc="upper right")
    axs[2].axvline(transit_result.period, alpha=0.4, lw=3)
    axs[2].set_xlim(np.min(tls_results.periods), np.max(tls_results.periods))
    for n in range(2, 10):
        axs[2].axvline(n * tls_results.period, alpha=0.4, lw=1, linestyle="dashed")
        axs[2].axvline(tls_results.period / n, alpha=0.4, lw=1, linestyle="dashed")
    axs[2].set(xlabel='Period (days)', ylabel='SDE')
    axs[2].plot(tls_results.periods, tls_results.power, color='black', lw=0.5)
    if plot_harmonics:
        max_harmonic_power_index = np.argmax(transit_result.harmonic_spectrum)
        harmonics = [1 / 4, 1 / 3, 1 / 2, 1, 2, 3, 4]
        for harmonic in harmonics:
            axs[3].axvline(harmonic * tls_results.periods[max_harmonic_power_index], alpha=0.4, lw=1, linestyle="dashed")
        axs[3].axvline(tls_results.periods[max_harmonic_power_index], alpha=0.4, lw=3)
        axs[3].set_xlim(np.min(tls_results.periods), np.max(tls_results.periods))
        axs[3].set(xlabel='Period (days)', ylabel='Harmonics Power')
        axs[3].plot(tls_results.periods, transit_result.harmonic_spectrum, color='black', lw=0.5)
    plt.savefig(plot_dir + file, bbox_inches='tight', dpi=200)
    fig.clf()
    plt.close(fig)



import multiprocessing
import os

from lcbuilder.objectinfo.ObjectInfo import ObjectInfo

from sherlockpipe.scoring.AverageSpectrumSignalSelector import AverageSpectrumSignalSelector
from sherlockpipe.scoring.BasicSdeSignalSelector import BasicSdeSignalSelector
from sherlockpipe.scoring.BasicSignalSelector import BasicSignalSelector
from sherlockpipe.scoring.QuorumSdeBorderCorrectedSignalSelector import QuorumSdeBorderCorrectedSignalSelector
from sherlockpipe.scoring.QuorumSnrBorderCorrectedSignalSelector import QuorumSnrBorderCorrectedSignalSelector
from sherlockpipe.scoring.SdeBorderCorrectedSignalSelector import SdeBorderCorrectedSignalSelector
from sherlockpipe.scoring.SnrBorderCorrectedSignalSelector import SnrBorderCorrectedSignalSelector
from sherlockpipe.search_zones.HabitableSearchZone import HabitableSearchZone
from sherlockpipe.search_zones.OptimisticHabitableSearchZone import OptimisticHabitableSearchZone


class SherlockTarget:
    """
    Used as input for :class:`sherlockpipe.sherlock.Sherlock`:
    """
    MASK_MODES = ['mask', 'subtract']
    VALID_SIGNAL_SELECTORS = ["basic", "border-correct", "quorum", "basic-snr", "border-correct-snr", "quorum-snr",
                              'average-spectrum']

    def __init__(self, object_info,
                 detrend_method='biweight', detrend_l_min=None, detrend_l_max=None, detrends_number=10,
                 detrend_cores=multiprocessing.cpu_count() - 1,
                 custom_selection_algorithm=None, custom_transit_template=None,
                 search_zone=None, custom_search_zone=None,
                 snr_min=5, sde_min=5,
                 min_sectors=0, max_sectors=99999,
                 bin_minutes=10,
                 mask_mode='mask',
                 cpu_cores=multiprocessing.cpu_count() - 1, max_runs=10, period_min=0.5,
                 period_max=33, period_protect=10, best_signal_algorithm='border-correct', quorum_strength=1,
                 min_quorum=0, fit_method='tls', oversampling=10,
                 t0_fit_margin=0.05, duration_grid_step=1.1,
                 source_properties_file=None,
                 cache_dir=os.path.expanduser('~') + "/",
                 ignore_original=False, pickle_mode='none', use_harmonics_spectra=False,
                 ois_mask=False):
        self.min_sectors = min_sectors
        self.max_sectors = max_sectors
        self.bin_minutes = bin_minutes
        self.mask_mode = mask_mode
        self.cpu_cores = cpu_cores if cpu_cores <= os.cpu_count() else os.cpu_count()
        self.max_runs = max_runs
        self.period_min = period_min
        self.period_max = period_max
        self.period_protect = period_protect
        self.best_signal_algorithm = best_signal_algorithm
        self.quorum_strength = quorum_strength
        self.min_quorum = min_quorum
        self.fit_method = fit_method
        self.oversampling = oversampling
        self.t0_fit_margin = t0_fit_margin
        self.duration_grid_step = duration_grid_step
        self.sde_min = sde_min
        self.snr_min = snr_min
        self.custom_search_zone = custom_search_zone
        self.search_zone = search_zone
        self.custom_transit_template = custom_transit_template
        self.custom_selection_algorithm = custom_selection_algorithm
        self.detrend_cores = detrend_cores if detrend_cores <= os.cpu_count() else os.cpu_count()
        self.detrends_number = detrends_number
        self.detrend_l_max = detrend_l_max
        self.detrend_l_min = detrend_l_min
        self.detrend_method = detrend_method
        self.pickle_mode = pickle_mode
        if mask_mode not in self.MASK_MODES:
            raise ValueError("Provided mask mode '" + mask_mode + "' is not allowed.")
        if best_signal_algorithm not in self.VALID_SIGNAL_SELECTORS:
            raise ValueError("Provided best signal algorithm '" + best_signal_algorithm + "' is not allowed.")
        self.search_zones_resolvers = {'hz': HabitableSearchZone(),
                                       'ohz': OptimisticHabitableSearchZone()}
        self.search_zone = search_zone if custom_search_zone is None else "user"
        if custom_search_zone is not None:
            self.search_zones_resolvers["user"] = custom_search_zone
        self.signal_score_selectors = {self.VALID_SIGNAL_SELECTORS[0]: BasicSdeSignalSelector(),
                                       self.VALID_SIGNAL_SELECTORS[1]: SdeBorderCorrectedSignalSelector(),
                                       self.VALID_SIGNAL_SELECTORS[2]: QuorumSdeBorderCorrectedSignalSelector(
                                           quorum_strength, min_quorum),
                                       self.VALID_SIGNAL_SELECTORS[3]: BasicSignalSelector(),
                                       self.VALID_SIGNAL_SELECTORS[4]: SnrBorderCorrectedSignalSelector(),
                                       self.VALID_SIGNAL_SELECTORS[5]: QuorumSnrBorderCorrectedSignalSelector(
                                           quorum_strength, min_quorum),
                                       self.VALID_SIGNAL_SELECTORS[6]: AverageSpectrumSignalSelector(),
                                       "user": custom_selection_algorithm}
        self.best_signal_algorithm = best_signal_algorithm if custom_selection_algorithm is None else "user"
        self.fit_method = "default"
        if fit_method is not None and fit_method.lower() == 'bls':
            self.fit_method = "box"
        elif fit_method is not None and fit_method.lower() == 'grazing':
            self.fit_method = "grazing"
        elif fit_method is not None and fit_method.lower() == 'tailed':
            self.fit_method = "tailed"
        self.oversampling = oversampling
        if self.oversampling is not None:
            self.oversampling = self.oversampling
        if custom_transit_template is not None:
            self.fit_method = "custom"
            self.user_transit_template = custom_transit_template
        self.object_info = object_info
        self.source_properties_file = source_properties_file
        self.cache_dir = cache_dir
        self.ignore_original = ignore_original
        self.use_harmonics_spectra = use_harmonics_spectra
        self.ois_mask = ois_mask



import datetime
from astropy.coordinates import Angle
from reportlab.lib import colors
from reportlab.lib.pagesizes import A4
from reportlab.lib.units import inch, cm
from reportlab.lib.enums import TA_CENTER, TA_JUSTIFY
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.platypus import BaseDocTemplate, PageTemplate, Frame, Paragraph, Spacer, Image, Table, TableStyle
from os import path
from uncertainties import ufloat
from astropy import units as u

width, height = A4
resources_dir = path.join(path.dirname(__file__))


class ObservationReport:
    """
    Used to create a pdf file from the parameters and images generated in the plan stage.
    """
    LOGO_IMAGE = resources_dir + "/../resources/images/sherlock3.png"
    ALERT_IMAGE = resources_dir + "/resources/images/alert.png"

    def __init__(self, df_observatories, df, alert_date, object_id, name, working_path, ra, dec, t0, t0_low_err, t0_up_err, period,
                 period_low_err, period_up_err, duration, duration_low_err, duration_up_err, depth, depth_low_err,
                 depth_up_err, observable, min_dist, max_dist, min_altitude, max_days, v, j, h, k):
        self.df = df
        self.df_observatories = df_observatories.drop('tz', 1)
        self.alert_date = alert_date
        self.object_id = object_id
        self.name = name
        self.working_path = working_path
        self.images_path = working_path + "/images/"
        self.observable = observable
        self.min_dist = min_dist
        self.max_dist = max_dist
        self.ra = ra
        self.dec = dec
        self.t0 = t0
        self.t0_low_err = t0_low_err
        self.t0_up_err = t0_up_err
        self.period = period
        self.period_low_err = period_low_err
        self.period_up_err = period_up_err
        self.duration = duration
        self.duration_low_err = duration_low_err
        self.duration_up_err = duration_up_err
        self.depth = depth
        self.depth_low_err = depth_low_err
        self.depth_up_err = depth_up_err
        self.min_altitude = min_altitude
        self.max_days = max_days
        self.v = v
        self.j = j
        self.h = h
        self.k = k

    def df_manipulations(self):
        """
        Performs changes of data from the initial dataframe for better formatting in the report.

        :return: the final dataframe
        """
        self.df['Observatory'] = self.df['observatory'].str.replace("-", " ")
        self.df['TZ'] = self.df['timezone'].fillna(0).astype('int')
        self.df['TZ'] = self.df['TZ'].astype('str')
        self.df['timezone'].astype(str)
        # Removing milliseconds from dates:
        self.df['ingress'] = self.df['ingress'].str[:-4]
        self.df['twilight_evening'] = self.df['twilight_evening'].str[:-4]
        self.df['midtime'] = self.df['midtime'].str[:-4]
        self.df['egress'] = self.df['egress'].str[:-4]
        self.df['twilight_morning'] = self.df['twilight_morning'].str[:-4]
        self.df['start_obs'] = self.df['start_obs'].str[:-4]
        self.df['end_obs'] = self.df['end_obs'].str[:-4]

        # Generamos la columna que contendrá el nombre de las imágenes:
        self.df['image_path'] = self.images_path + self.df['observatory'] + "_" + self.df['midtime'] + ".png"
        # Finalmente con las rutas creamos los objetos:
        self.df['Image'] = self.df['image_path'].apply(lambda x: Image(x))


        # We want to sort the dates row-wise, only keeping the interesting columns
        df_dates = self.df[['twilight_evening', 'ingress', 'midtime', 'egress', 'start_obs', 'end_obs', 'twilight_morning']]
        df_dates.columns = ['TWE', 'I', 'M', 'E', 'SO', 'EO', 'TWM']
        dict_fechas = df_dates.to_dict('index')
        new_dict_fechas = {}
        for item in dict_fechas:
            # Lo ordenamos por los valores de cada key:
            list_of_tuples = sorted(dict_fechas[item].items(), key=lambda item: item[1])
            # Definimos la noche como false:
            night = 0
            # Recorremos las tuplas y si noche es uno (lo que sucede a partir del TWE) ponemos en negrita la fecha.
            # Con esto vamos a generar una nueva lista de tuplas que tendrá las fechas apropiadas en negrita:
            list_of_tuples_with_night = []
            for tup in list_of_tuples:
                if night == 1 and tup[0] != 'TWM':
                    y = list(tup)
                    y[1] = '<strong>' + str(tup[1]) + '</strong>'
                    tup = tuple(y)
                if tup[0] == 'TWM':
                    night = 0
                if tup[0] == 'TWE':
                    night = 1
                list_of_tuples_with_night.append(tup)
            # Unimos las tuplas separando key y value con ":":
            join_list = [': '.join(tup) for tup in list_of_tuples_with_night]
            # Para terminar, el nuevo diccionario une los elementos de la lista separando por saltos de línea:
            new_dict_fechas[item] = '<br/>'.join(join_list)
        # Finalmente, creamos la columna Event times mapeando el index del dataframe con el del diccionario:
        self.df['Event times'] = self.df.index.to_series().map(new_dict_fechas)
        # La columna Transit Times Error será el concatenado de los errores:
        self.df['TT Error'] = '-' + self.df['midtime_low_err_h'].map(str) + '<br/>+' + self.df['midtime_up_err_h'].map(
            str)
        # La columna Moon será el concatenado de la moon_phase y de moon_dist:
        self.df['moon_phase'] = (self.df['moon_phase'] * 100).astype(int)
        self.df['moon_dist'] = self.df['moon_dist'].astype(int)
        self.df['Moon'] = self.df['moon_phase'].map(str) + '%<br/>' + self.df['moon_dist'].map(str) + 'º'
        # El dataframe final solo tendrá unas pocas columnas del excel inicial:
        df_output = self.df[['Observatory', 'TZ', 'Event times', 'TT Error', 'Moon', 'Image']]
        return df_output

    @staticmethod
    def row_colors(df, table_object):
        data_len = len(df)
        for each in range(1, data_len + 1):
            if each % 2 == 1:
                bg_color = colors.whitesmoke
            else:
                bg_color = colors.lightgrey

            table_object.setStyle(TableStyle([('BACKGROUND', (0, each), (-1, each), bg_color)]))

    def create_header(self, canvas, doc):
        """
        Initializes the common header for all the pages.

        :param canvas: the report canvas
        :param doc: the report document
        """
        canvas.saveState()
        canvas.drawImage(self.LOGO_IMAGE, x=0, y=26.8 * cm, height=2.7 * cm, width=2.7 * cm, preserveAspectRatio=True)
        object_id_text = 'Sherlock observation plan: %s' % self.object_id
        canvas.setFont(psfontname="Helvetica", size=12)
        canvas.drawRightString(x=12 * cm, y=28 * cm, text=object_id_text)
        if doc.page == 1:
            object_id_text = '%s OBSERVATION PLAN' % self.object_id
            canvas.setFont(psfontname="Helvetica-Bold", size=25)
            canvas.drawCentredString(x=10 * cm, y=25.5 * cm, text=object_id_text)
        report_date = datetime.datetime.now().strftime("%a, %d %B %Y, %H:%M:%S")
        report_date_text = '%s' % report_date
        canvas.setFont("Helvetica", 9)
        canvas.drawRightString(20.5 * cm, 28 * cm, report_date_text)
        canvas.restoreState()

    def create_footer(self, canvas, doc):
        """
        Initializes the common footer for all the pages

        :param canvas: the report canvas
        :param doc: the report document
        """
        canvas.saveState()
        # if doc.page == 1:
        #     # Footer con superíndice:
        #     textobject = canvas.beginText()
        #     textobject.setTextOrigin(1.8 * cm, 2.1 * cm)
        #     textobject.setFont("Helvetica", 5)
        #     textobject.setRise(5)
        #     textobject.textOut('1 ')
        #     textobject.setRise(0)
        #     textobject.setFont("Helvetica", 7)
        #     pie_pagina = 'Three possible observability values are defined: 1 - Entire transit is required, ' \
        #                  '0.5 - Transit midtime and either ingress or egress at least are required,\n' \
        #                  '0.25 - Only ingress or egress are required, with moon constraints of % sº as minimum ' \
        #                  'distance for new moon and % sº as minimum distance for full moon\n' \
        #                  'and for the observatories listed in the Table 2.' % (self.min_dist, self.max_dist)
        #
        #     for line in pie_pagina.splitlines():
        #         textobject.textLine(line)
        #
        #     canvas.drawText(textobject)
        page = "Powered by Astropy, Astroplan and ReportLab"
        canvas.setFont("Helvetica", 9)
        canvas.drawRightString(7 * cm, 0.5 * cm, page)
        page = "Page %s" % doc.page
        canvas.setFont("Helvetica", 9)
        canvas.drawRightString(20.5 * cm, 0.5 * cm, page)
        canvas.restoreState()

    def create_report(self):
        """
        Creates the final report with all the star parameters, the candidate information and the observation nights
        data, storing it in a pdf.
        """
        df_manipulated = self.df_manipulations()
        # Styles to be used
        styles = getSampleStyleSheet()
        styles.add(ParagraphStyle(name="ParagraphAlignCenter", alignment=TA_CENTER))
        styles.add(ParagraphStyle(name="ParagraphAlignJustify", alignment=TA_JUSTIFY))
        styles.wordWrap = 'LTR'
        table_style = TableStyle([('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),
                                  ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
                                  ('INNERGRID', (0, 0), (-1, -1), 0.25, colors.black),
                                  ('BOX', (0, 0), (-1, -1), 0.25, colors.black),
                                  ('FONTSIZE', (0, 0), (-1, -1), 10),
                                  ])

        # Content:
        story = [Spacer(1, 75)]
        introduction = '<font name="HELVETICA" size="9">This document is generated by the SHERLOCK PIPEline \
        transiting candidates observation planner. The target star this document focuses on is the %s. The star \
        parameters are shown in Table 1. The proposed candidate parameters are those exposed in Table 1. The  \
        observables included in the Table 5 are obtained based on the constraints of Table 3 and the list of \
        observatories from Table 4.</font>' % self.object_id
        story.append(Paragraph(introduction, styles["ParagraphAlignJustify"]))

        story.append(Spacer(1, 30))

        # Generamos la tabla 1 con los parámetros:
        tabla1_data = [['RA (deg)', 'Dec (deg)', 'V (mag)', 'J (mag)', 'H (mag)', 'K (mag)'],
                       [Angle(self.ra, u.deg).to_string(unit=u.hourangle, sep=':', precision=2),
                        Angle(self.dec, u.deg).to_string(unit=u.deg, sep=':', precision=2),
                        round(self.v, 2), round(self.j, 2), round(self.h, 2), round(self.k, 2)]]
        table1_colwidth = [3.5 * cm, 3.5 * cm, 2 * cm, 2 * cm, 2 * cm, 2 * cm]
        table1_number_rows = len(tabla1_data)
        tabla1 = Table(tabla1_data, table1_colwidth, table1_number_rows * [0.75 * cm])
        tabla1.setStyle(table_style)
        # Le damos el estilo alternando colores de filas:
        ObservationReport.row_colors(tabla1_data, tabla1)
        story.append(tabla1)

        table1_descripcion = '<font name="HELVETICA" size="9"><strong>Table 1: </strong>\
                        The proposed target parameters.</font>'
        story.append(Spacer(1, 5))
        story.append(Paragraph(table1_descripcion, styles["ParagraphAlignCenter"]))

        story.append(Spacer(1, 30))

        # Generamos la tabla 2 con los parámetros:
        tabla2_data = [['T0 (TBJD)', 'Period (d)', 'Duration (h)', 'Depth (ppt)'],
                        [ufloat(self.t0, self.t0_low_err, self.t0_up_err),
                        ufloat(self.period, self.period_low_err, self.period_up_err),
                        ufloat(self.duration, self.duration_low_err, self.duration_up_err),
                        ufloat(self.depth, self.depth_low_err, self.depth_up_err)]]
        table2_colwidth = [4 * cm, 4 * cm, 3.5 * cm, 3.5 * cm]
        table2_number_rows = len(tabla2_data)
        tabla2 = Table(tabla2_data, table2_colwidth, table2_number_rows * [0.75 * cm])
        tabla2.setStyle(table_style)
        # Le damos el estilo alternando colores de filas:
        ObservationReport.row_colors(tabla2_data, tabla2)
        story.append(tabla2)

        table1_descripcion = '<font name="HELVETICA" size="9"><strong>Table 2: </strong>\
                The proposed candidate parameters.</font>'
        story.append(Spacer(1, 5))
        story.append(Paragraph(table1_descripcion, styles["ParagraphAlignCenter"]))

        story.append(Spacer(1, 30))

        # Generamos la tabla 3 con los parámetros:
        tabla3_data = [['Observability', 'Days interval', 'Min altitude (º)', 'Min Moon Dist (º)', 'Max Moon Dist (º)'],
                       [self.observable, self.max_days, self.min_altitude, self.min_dist, self.max_dist]]
        table3_colwidth = [2.5 * cm, 2.5 * cm, 2.5 * cm, 3 * cm, 3 * cm]
        table3_number_rows = len(tabla3_data)
        tabla3 = Table(tabla3_data, table3_colwidth, table3_number_rows * [0.75 * cm])
        tabla3.setStyle(table_style)
        # Le damos el estilo alternando colores de filas:
        ObservationReport.row_colors(tabla3_data, tabla3)
        story.append(tabla3)

        table3_descripcion = '<font name="HELVETICA" size="9"><strong>Table 3: </strong>\
                        The constraints used to calculate the observables. Observability can have the next values: ' \
                             '1 - Entire transit is required, ' \
                             '0.5 - Transit midtime and either ingress or egress at least are required,\n' \
                             '0.25 - Only ingress or egress are required. ' \
                             'Days interval is the maximum number of days to search for observable transits.' \
                             'Min altitude is the minimum altitude above the horizon to consider an observable ' \
                             'transit. Min Moon Dist is the minimum distance to be kept from the target to the new ' \
                             'moon. Max Moon Dist is the minimum distance to be kept from the target to the full ' \
                             'moon. </font>'
        story.append(Spacer(1, 5))
        story.append(Paragraph(table3_descripcion, styles["ParagraphAlignCenter"]))

        story.append(Spacer(1, 30))

        # Generamos la tabla 4 con los observatorios:
        table4_colwidth = [3.5 * cm, 2.5 * cm, 2.5 * cm, 2.5 * cm]
        table4_number_rows = len(self.df_observatories) + 1  # Sumamos 1 para tener en cuenta el header
        table4_header_row = [['Name', 'Latitude (º)', 'Longitude (º)', 'Altitude (m)']]
        table4_data = self.df_observatories.values.tolist()
        tabla4 = Table(table4_header_row + table4_data, table4_colwidth, table4_number_rows * [0.75 * cm])
        tabla4.setStyle(table_style)
        # Le damos el estilo alternando colores de filas:
        ObservationReport.row_colors(self.df_observatories, tabla4)
        story.append(tabla4)

        table4_descripcion = '<font name="HELVETICA" size="9"><strong>Table 3: </strong>\
                List of observatories.</font>'
        story.append(Spacer(1, 5))
        story.append(Paragraph(table4_descripcion, styles["ParagraphAlignCenter"]))

        if self.alert_date is not None:
            story.append(Spacer(1, 30))
            story.append(Image(self.ALERT_IMAGE, width=15, height=15))
            alert = '<font name="HELVETICA" size="9" color="red">The table 5 with the observations will only contain \
                    scheduled observation nights up to the %s. Since then, the error propagation gets too wide to be \
                    acceptable for the plan to be precise enough.</font>' % self.alert_date
            story.append(Paragraph(alert, styles["ParagraphAlignJustify"]))

        # Pasamos a la siguiente página:
        story.append(Spacer(1, 30))

        # Generamos la tabla 5 con los observables:
        table5_header_row = [df_manipulated.columns[:, ].values.astype(str).tolist()]
        table5_data = df_manipulated.values.tolist()
        table5_data2 = []
        for row in table5_data:
            row_counter = 1
            manipulated_data = []
            for cell in row:
                # La última de las columnas se corresponde con imágenes, la mantenemos:
                if row_counter == len(row):
                    manipulated_data.append(cell)
                else:
                    manipulated_data.append(Paragraph(cell, styles["ParagraphAlignCenter"]))
                row_counter = row_counter + 1
            table5_data2.append(manipulated_data)

        final_data = table5_header_row + table5_data2

        # Creates a table with variable number of row and width:
        table5_number_rows = len(final_data)
        table5_rowwidths = []
        for x in range(table5_number_rows):
            if x == 0:  # Table headers
                table5_rowwidths.append(0.2 * inch)
            else:
                table5_rowwidths.append(1.2 * inch)

        # Creates a table with 5 columns, variable width:
        table5_colwidths = [1.1 * inch, 0.3 * inch, 2.3 * inch, 0.7 * inch, 0.7 * inch, 2.2 * inch]
        tabla5 = Table(final_data, table5_colwidths, table5_rowwidths, repeatRows=1)
        tabla5.setStyle(table_style)
        # Le damos el estilo alternando colores de filas:
        ObservationReport.row_colors(final_data, tabla5)

        story.append(tabla5)

        table5_descripcion = '<font name="HELVETICA" size="9"><strong>Table 5</strong>: Observables for the \
                computed candidate. <strong>TZ column</strong> represents the observatory time zone offset from \
                UTC for each transit event. <strong>TT errors column</strong> represents the uncertainty of the \
                ingress, midtime and egress times for each observable calculated from the T0 and Period uncertainties. \
                <strong>Moon column</strong> represents the moon phase and distance to the target star at the \
                time of the transit midtime. <strong>Image column</strong> represents a plot where the altitude \
                and air mass are plotted with a blue line. The transit midtime is plotted as a vertical black bar \
                together with the ingress and egress in orange lines. The red vertical bars represent the uncertainty \
                of the ingress (left) and the egress (right). The green floor of the graph represents \
                the limit of altitude / air mass for the event to be observable. The white background and the \
                gray background represent the daylight and night based on the nautical twilights.</font>'
        story.append(Spacer(1, 5))
        story.append(Paragraph(table5_descripcion, styles["ParagraphAlignJustify"]))
        # Definimos el frame y el template sabiendo que un A4: 21 cm de ancho por 29,7 cm de altura
        global_frame = Frame(1.5 * cm, 1.1 * cm, 18 * cm, 25.4 * cm, id='normal', showBoundary=0)
        global_template = PageTemplate(id='UnaColumna', frames=global_frame,
                                       onPage=self.create_header, onPageEnd=self.create_footer)
        # Construimos el documento:
        doc = BaseDocTemplate(self.working_path + "/" + self.name + "_observation_plan.pdf", pagesize=A4,
                              rightMargin=40, leftMargin=40,
                              topMargin=95, bottomMargin=15,
                              pageTemplates=global_template)
        doc.build(story)



from watson.watson import Watson

from sherlockpipe.loading.tool_with_candidate import ToolWithCandidate


class Vetter(ToolWithCandidate):
    watson = None

    def __init__(self, object_dir, is_candidate_from_search, candidates_df) -> None:
        super().__init__(is_candidate_from_search, candidates_df)
        self.watson = Watson(object_dir)

    def run(self, cpus, **kwargs):
        self.watson.vetting_with_data(kwargs['candidate'], kwargs['star_df'], kwargs['transits_df'],
                                      cpus, transits_mask=kwargs["transits_mask"])

    def object_dir(self):
        return self.watson.object_dir



import dataclasses
import logging
import math
import re
import shutil
from pathlib import Path

import alexfitter
import numpy as np
import pandas as pd
import os
from os import path
import matplotlib.pyplot as plt
from lcbuilder.helper import LcbuilderHelper

from sherlockpipe.loading.tool_with_candidate import ToolWithCandidate


resources_dir = path.join(path.dirname(__file__))
resources_dir = str(Path(resources_dir).parent.absolute())


@dataclasses.dataclass
class DistributionParams:
    yerr: float = -7.0
    yerr_lower_err: float = -15.0
    yerr_upper_err: float = 0.0
    lnsigma: float = 0.0
    lnsigma_lower_err: float = -30.0
    lnsigma_upper_err: float = 30.0
    lnrho: float = 0.0
    lnrho_lower_err: float = -30.0
    lnrho_upper_err: float = 30.0


class Fitter2(ToolWithCandidate):
    """
    Used to run a bayesian fit on the Sherlock search results.
    """

    def __init__(self, object_dir, fit_dir, only_initial, is_candidate_from_search, candidates_df, mcmc=False,
                 detrend=None, estimate_noise=True, rho_err_multi=5.0, sigma_err_multi=5.0, yerr_err_multi=5.0):
        super().__init__(is_candidate_from_search, candidates_df)
        self.object_dir = os.getcwd() if object_dir is None else object_dir
        self.data_dir = fit_dir
        self.only_initial = only_initial
        self.mcmc = mcmc
        self.detrend = detrend
        self.estimate_noise = estimate_noise
        self.rho_err_multi = rho_err_multi
        self.sigma_err_multi = sigma_err_multi
        self.yerr_err_multi = yerr_err_multi

    def mask_non_fit_candidates(self, time, flux, flux_err, candidate_df, fit_candidate_df):
        """
        Masks all the candidates found in previous runs in the SHERLOCK search.

        :param time: the time array
        :param flux: the flux measurements array
        :param flux_err: the flux error measurements array
        :param candidate_df: the candidates used for noise estimation
        :param fit_candidate_df: the candidates used for fit
        :return: time, flux and flux_err with previous candidates in-transit data masked
        """
        mask_candidate_df = candidate_df.loc[~candidate_df['number'].isin(fit_candidate_df['number'])]
        for index, candidate_row in mask_candidate_df.iterrows():
            period = candidate_row["period"]
            duration = candidate_row["duration"]
            duration = duration / 60 / 24
            t0 = candidate_row["t0"]
            logging.info("Masking candidate number %.0f with P=%.3fd, T0=%.2f and D=%.2fd", index + 1, period, t0,
                         duration)
            time, flux, flux_err = LcbuilderHelper.mask_transits(time, flux, period, duration * 2, t0, flux_err)
        return time, flux, flux_err

    def fit(self, candidate_df, fit_candidate_df, star_df, cpus, allesfit_dir, tolerance, fit_orbit):
        """
        Main method to run the alexfitter fit.

        :param candidate_df: the candidates dataframe to be used for noise estimation.
        :param fit_candidate_df: the candidates dataframe to be used for the final fit.
        :param star_df: the star dataframe.
        :param cpus: the number of cpus to be used.
        :param allesfit_dir: the directory for the fit to be run.
        :param tolerance: the nested sampling tolerance threshold.
        :param fit_orbit: whether to fit eccentricity and arg. of periastron.
        """
        logging.info("Preparing fit files")
        sherlock_star_file = self.object_dir + "/params_star.csv"
        star_file = allesfit_dir + "/params_star.csv"
        params_file = allesfit_dir + "/params.csv"
        settings_file = allesfit_dir + "/settings.csv"
        run = int(candidate_df['number'].iloc[0])
        # We load the unprocessed raw PDCSAP flux curve
        lc = pd.read_csv(self.object_dir + '/lc.csv', header=0)
        time, flux, flux_err = lc["#time"].values, lc["flux"].values, lc["flux_err"].values
        #time, flux, flux_err = self.mask_previous_candidates(time, flux, flux_err, run)
        lc = pd.DataFrame(columns=['#time', 'flux', 'flux_err'])
        lc['#time'] = time
        lc['flux'] = flux
        lc['flux_err'] = flux_err
        lc.to_csv(allesfit_dir + "/lc.csv", index=False)
        if os.path.exists(sherlock_star_file) and os.path.isfile(sherlock_star_file):
            shutil.copyfile(sherlock_star_file, star_file)
        logging.info("Preparing fit properties")
        self.overwrite_settings(settings_file, cpus, candidate_df, tolerance, "multi")
        with open(params_file, 'w') as f:
            f.write(self.fill_candidates_params(candidate_df, star_df, fit_orbit, self.detrend))
            f.truncate()
        logging.info("Running initial guess")
        try:
            alexfitter.show_initial_guess(allesfit_dir)
        except Exception as e:
            logging.exception(str(e))
        # TODO fix custom_plot for all candidates
        # Fitter.custom_plot(candidate_row["name"], candidate_row["period"], fit_width, allesfit_dir, "initial_guess")
        shutil.copytree(allesfit_dir + '/results', allesfit_dir + '/results_initial_guess_before_noise')
        shutil.copy(allesfit_dir + '/params.csv', allesfit_dir + '/params_before_noise.csv')
        shutil.copy(allesfit_dir + '/settings.csv', allesfit_dir + '/settings_before_noise.csv')
        if not self.only_initial:
            logging.info("Preparing bayesian fit")
            if self.estimate_noise and self.detrend == 'gp':
                logging.info("Running noise estimation")
                alexfitter.estimate_noise_out_of_transit(allesfit_dir)
                noise_estimation = pd.read_csv(allesfit_dir + "/priors/summary_phot.csv")
                noise_estimation = noise_estimation.iloc[0]
                noise_distribution_params = DistributionParams(yerr=noise_estimation['ln_yerr_median'],
                                   yerr_lower_err=noise_estimation['ln_yerr_ll'],
                                   yerr_upper_err=noise_estimation['ln_yerr_ul'],
                                   lnsigma=noise_estimation['gp_ln_sigma_median'],
                                   lnsigma_lower_err=noise_estimation['gp_ln_sigma_ll'],
                                   lnsigma_upper_err=noise_estimation['gp_ln_sigma_ul'],
                                   lnrho=noise_estimation['gp_ln_rho_median'],
                                   lnrho_lower_err=noise_estimation['gp_ln_rho_ll'],
                                   lnrho_upper_err=noise_estimation['gp_ln_rho_ul'])
                self.overwrite_settings(settings_file, cpus, fit_candidate_df, tolerance, "multi")
                with open(params_file, 'w') as f:
                    f.write(self.fill_candidates_params(fit_candidate_df, star_df, fit_orbit, self.detrend,
                                                           distribution='normal',
                                                           distribution_params=noise_distribution_params))
                    f.truncate()
            time, flux, flux_err = self.mask_non_fit_candidates(time, flux, flux_err, candidate_df, fit_candidate_df)
            # time, flux, flux_err = self.mask_previous_candidates(time, flux, flux_err, run)
            lc = pd.DataFrame(columns=['#time', 'flux', 'flux_err'])
            lc['#time'] = time
            lc['flux'] = flux
            lc['flux_err'] = flux_err
            lc.to_csv(allesfit_dir + "/lc.csv", index=False)
            if not self.mcmc:
                logging.info("Running dynamic nested sampling")
                try:
                    alexfitter.show_initial_guess(allesfit_dir)
                    alexfitter.ns_fit(allesfit_dir)
                    alexfitter.ns_output(allesfit_dir)
                except Exception as e:
                    logging.exception(str(e))
            elif self.mcmc:
                logging.info("Running MCMC")
                try:
                    alexfitter.mcmc_fit(allesfit_dir)
                    alexfitter.mcmc_output(allesfit_dir)
                except Exception as e:
                    logging.exception(str(e))
            logging.info("Generating custom plots")
            # TODO fix custom_plot for all candidates
            # Fitter.custom_plot(candidate_row["name"], candidate_row["period"], fit_width, allesfit_dir, "posterior")


    def overwrite_settings(self, settings_file, cpus, candidate_df, tolerance, boundaries="single"):
        shutil.copyfile(resources_dir + "/resources/allesfitter/settings2.csv", settings_file)
        fit_width = Fitter2.select_fit_width(candidate_df)
        with open(settings_file, 'r+') as f:
            text = f.read()
            text = re.sub('\\${sherlock:cores}', str(cpus), text)
            text = re.sub('\\${sherlock:fit_width}', str(fit_width), text)
            text = re.sub('\\${sherlock:fit_ttvs}', "False", text)
            text = re.sub('\\${sherlock:names}', ' '.join(candidate_df["name"].astype('str')), text)
            text = re.sub('\\${sherlock:boundaries}', str(boundaries), text)
            text = re.sub('\\${sherlock:tolerance}', str(tolerance), text)
            if self.detrend == 'hybrid_spline':
                detrend_param = "baseline_flux_lc,hybrid_spline"
            elif self.detrend == 'gp':
                detrend_param = 'baseline_flux_lc,sample_GP_Matern32'
            else:
                detrend_param = ''
            text = re.sub('\\${sherlock:detrend}', detrend_param, text)
            f.seek(0)
            f.write(text)
            f.truncate()

    @staticmethod
    def select_fit_width(candidate_df):
        """
        Calculates the window to be used for the fit around the transits.

        :param candidate_df: the candidates dataframe
        :return: the maximum calculated fit_width
        """
        fit_width = 0
        for key, row in candidate_df.iterrows():
            if row["duration"] is not None:
                row_fit_width = float(row["duration"]) / 60 / 24 * 7
                fit_width = fit_width if fit_width > row_fit_width else row_fit_width
        return fit_width

    @staticmethod
    def custom_plot(name, period, fit_width, allesfit_dir, mode="posterior"):
        """
        Creates a custom fit plot from the allesfitter data

        :param name: the candidate name
        :param period: the final period
        :param fit_width: the fit_width window
        :param allesfit_dir: the directory where allesfitter data is stred
        :param mode: the allesfitter plot model
        """
        allesclass = alexfitter.allesclass(allesfit_dir)
        baseline_width = fit_width * 24
        baseline_to_period = fit_width / period
        fig, axes = plt.subplots(2, 3, figsize=(18, 6), gridspec_kw={'height_ratios': [3, 1]}, sharex='col')
        fig.subplots_adjust(left=0.5, right=1.5, hspace=0)
        style = 'full'
        allesclass.plot('lc', name, style, ax=axes[0][0], mode=mode)
        axes[0][0].set_title('lc, ' + style)
        allesclass.plot('lc', name, style + '_residuals', ax=axes[1][0], mode=mode)
        axes[1][0].set_title('')
        style = 'phase'
        allesclass.plot('lc', name, style, ax=axes[0][1], mode=mode, zoomwindow=baseline_to_period)
        axes[0][1].set_title('lc, ' + style)
        axes[0][1].set_xlim([- baseline_to_period / 2, baseline_to_period / 2])
        allesclass.plot('lc', name, style + '_residuals', ax=axes[1][1], mode=mode, zoomwindow=baseline_to_period)
        axes[1][1].set_title('')
        axes[1][1].set_xlim([- baseline_to_period / 2, baseline_to_period / 2])
        style = 'phasezoom'
        allesclass.plot('lc', name, style, ax=axes[0][2], mode=mode, zoomwindow=baseline_width)
        axes[0][2].set_title('lc, ' + style)
        axes[0][2].set_xlim([- baseline_width / 2, baseline_width / 2])
        allesclass.plot('lc', name, style + '_residuals', ax=axes[1][2], mode=mode, zoomwindow=baseline_width)
        axes[1][2].set_title('')
        axes[1][2].set_xlim([- baseline_width / 2, baseline_width / 2])
        fig.savefig(allesfit_dir + '/results/ns_' + mode + '_' + name + '_custom.pdf', bbox_inches='tight')
        style = ['phasezoom_occ']
        style = ['phase_variations']

    def fill_candidates_params(self, candidate_df, star_df, fit_orbit, detrend_mode, distribution: str = 'uniform',
                               distribution_params: DistributionParams=None):
        """
        Fills the candidate planets initial parameters and their distribution to be fit.

        :param candidate_df: the candidates dataframe
        :param star_df: the star dataframe
        :param fit_orbit: whether to fit eccentricity and arg. of periastron.
        :param detrend_mode: type of detrend to be used
        :param distribution: uniform or normal
        :param distribution_params: the distribution parameters if known
        :return: the allesfitter candidates parameters
        """
        candidate_priors_text = ""
        for key, row in candidate_df.iterrows():
            candidate_priors_text = candidate_priors_text + \
                                    Fitter2.fill_candidate_params(row, star_df, fit_orbit)
        return candidate_priors_text + self.fill_instrument_params(star_df, detrend_mode, distribution,
                                                                      distribution_params)

    @staticmethod
    def fill_candidate_params(candidate_row, star_df, fit_orbit):
        """
        Fills the candidate planet initial parameters and their distribution to be fit.

        :param candidate_row: the candidate row from the dataframe
        :param star_df: the star dataframe
        :param fit_orbit: whether to fit eccentricity and arg. of periastron.
        :return: the allesfitter candidate parameters
        """
        candidate_params = """#name,value,fit,bounds,label,unit
#companion b astrophysical params,,,,,
${sherlock:name}_rr,${sherlock:rp_rs},1,uniform ${sherlock:rp_rs_min} ${sherlock:rp_rs_max},$R_b / R_\star$,
${sherlock:name}_rsuma,${sherlock:sum_rp_rs_a},1,uniform ${sherlock:sum_rp_rs_a_min} ${sherlock:sum_rp_rs_a_max},$(R_\star + R_b) / a_b$,
${sherlock:name}_cosi,0.0,1,uniform 0.0 0.06,$\cos{i_b}$,
${sherlock:name}_epoch,${sherlock:t0},1,uniform ${sherlock:t0_min} ${sherlock:t0_max},$T_{0;b}$,$\mathrm{BJD}$
${sherlock:name}_period,${sherlock:period},1,uniform ${sherlock:period_min} ${sherlock:period_max},$P_b$,$\mathrm{d}$
${sherlock:name}_f_c,0.0,${sherlock:fit_orbit},uniform -1.0 1.0,$\sqrt{e_b} \cos{\omega_b}$,
${sherlock:name}_f_s,0.0,${sherlock:fit_orbit},uniform -1.0 1.0,$\sqrt{e_b} \sin{\omega_b}$,
"""
        candidate_params = re.sub('\\${sherlock:t0}', str(candidate_row["t0"]), candidate_params)
        candidate_params = re.sub('\\${sherlock:t0_min}', str(candidate_row["t0"] - 0.02), candidate_params)
        candidate_params = re.sub('\\${sherlock:t0_max}', str(candidate_row["t0"] + 0.02), candidate_params)
        candidate_params = re.sub('\\${sherlock:period}', str(candidate_row["period"]), candidate_params)
        candidate_params = re.sub('\\${sherlock:period_min}', str(candidate_row["period"] - candidate_row["per_err"]),
                                  candidate_params)
        candidate_params = re.sub('\\${sherlock:period_max}', str(candidate_row["period"] + candidate_row["per_err"]),
                                  candidate_params)
        candidate_params = re.sub('\\${sherlock:fit_orbit}', str(int(fit_orbit)), candidate_params)
        rp_rs = candidate_row["rp_rs"] if candidate_row["rp_rs"] != "-" else 0.1
        depth = candidate_row["depth"] / 1000
        # TODO calculate depth error in SHERLOCK maybe given the std deviation from the depths or even using the
        #  residuals
        depth_err = depth * 0.2
        rp_rs_err = 0.5 / math.sqrt(depth) * depth_err
        candidate_params = re.sub('\\${sherlock:rp_rs}', str(rp_rs), candidate_params)
        rp_rs_min = rp_rs - 2 * rp_rs_err
        rp_rs_min = rp_rs_min if rp_rs_min > 0 else 0.0000001
        rp_rs_max = rp_rs + 2 * rp_rs_err
        rp_rs_max = rp_rs_max if rp_rs_max < 1 else 0.9999
        candidate_params = re.sub('\\${sherlock:rp_rs_min}', str(rp_rs_min), candidate_params)
        candidate_params = re.sub('\\${sherlock:rp_rs_max}', str(rp_rs_max), candidate_params)
        G = 6.674e-11
        mstar_kg = star_df.iloc[0]["M_star"] * 2e30
        mstar_kg_lerr = star_df.iloc[0]["M_star_lerr"] * 2e30
        mstar_kg_uerr = star_df.iloc[0]["M_star_uerr"] * 2e30
        rstar_au = star_df.iloc[0]['R_star'] * 0.00465047
        rstar_lerr_au = star_df.iloc[0]["R_star_lerr"] * 0.00465047
        rstar_uerr_au = star_df.iloc[0]["R_star_uerr"] * 0.00465047
        per = candidate_row["period"] * 24 * 3600
        per_err = candidate_row["per_err"] * 24 * 3600
        a = (G * mstar_kg * per ** 2 / 4. / (np.pi ** 2)) ** (1. / 3.)
        a_au = a / 1.496e11
        sum_rp_rs_a = (np.sqrt(depth) + 1.) * rstar_au / a_au
        sum_rp_rs_a_lerr = np.sqrt(
            np.square(0.5 * depth_err / depth) + np.square(rstar_lerr_au / rstar_au) + np.square(
                2 * per_err / 3. / per) + np.square(mstar_kg_lerr / 3. / mstar_kg)) * sum_rp_rs_a
        sum_rp_rs_a_uerr = np.sqrt(
            np.square(0.5 * depth_err / depth) + np.square(rstar_uerr_au / rstar_au) + np.square(
                2 * per_err / 3. / per) + np.square(mstar_kg_uerr / 3. / mstar_kg)) * sum_rp_rs_a
        sum_rp_rs_a_min = sum_rp_rs_a - 2 * sum_rp_rs_a_lerr
        sum_rp_rs_a_max = sum_rp_rs_a + 2 * sum_rp_rs_a_uerr
        sum_rp_rs_a_min = sum_rp_rs_a_min if sum_rp_rs_a_min > 0 else 0.0000001
        candidate_params = re.sub('\\${sherlock:sum_rp_rs_a}', str(sum_rp_rs_a), candidate_params)
        candidate_params = re.sub('\\${sherlock:sum_rp_rs_a_min}', str(sum_rp_rs_a_min), candidate_params)
        candidate_params = re.sub('\\${sherlock:sum_rp_rs_a_max}', str(sum_rp_rs_a_max), candidate_params)
        candidate_params = re.sub('\\${sherlock:name}', str(candidate_row["name"]), candidate_params)
        return candidate_params
    def fill_instrument_params(self, star_df, detrend_mode, distribution='uniform', distribution_params: DistributionParams=None):
        """
        Fills the star and systematics information for each instrument (so far only one)

        :param star_df: the star dataframe
        :param detrend_mode: type of detrend to be used
        :param distribution: distribution: uniform or normal
        :param distribution_params : if set, chooses the baseline fit values
        :return: the allesfitter instrument params
        """
        instrument_params = """#limb darkening coefficients per instrument,,,,,
host_ldc_q1_lc,${sherlock:ld_a},1,uniform 0.0 1.0,$q_{1; \mathrm{lc}}$,
host_ldc_q2_lc,${sherlock:ld_b},1,uniform 0.0 1.0,$q_{2; \mathrm{lc}}$,
#spots per instrument and companion,,,,,
#errors per instrument,
${sherlock:err_params}
#baseline per instrument,
${sherlock:baseline_params}
"""
        if star_df.iloc[0]["ld_a"] is not None and not np.isnan(star_df.iloc[0]["ld_a"]):
            instrument_params = re.sub('\\${sherlock:ld_a}', str(star_df.iloc[0]["ld_a"]), instrument_params)
            instrument_params = re.sub('\\${sherlock:ld_b}', str(star_df.iloc[0]["ld_b"]), instrument_params)
        else:
            instrument_params = re.sub('\\${sherlock:ld_a}', "0.5", instrument_params)
            instrument_params = re.sub('\\${sherlock:ld_b}', "0.5", instrument_params)
        if detrend_mode == 'gp':
            if distribution_params is None:
                distribution_params = DistributionParams()
            if 'uniform' == distribution:
                errs_params = "ln_err_flux_lc," + str(distribution_params.yerr) + ",1," + distribution + " " + \
                              str(distribution_params.yerr_lower_err) + " " + \
                              str(distribution_params.yerr_upper_err) + "," \
                              "$\\\\log{\\\\sigma_\\\\mathrm{lc}}$,\\$\\\\log{ \\\\mathrm{rel. flux.} }$"
                baseline_params = "baseline_gp_matern32_lnsigma_flux_lc," + str(
                    distribution_params.lnsigma) + ",1," + distribution + \
                                  " " + str(distribution_params.lnsigma_lower_err) + " " + \
                                  str(distribution_params.lnsigma_upper_err) + "," \
                                  "$\\\\mathrm{gp: \\\\ln{\\\\sigma} (lc)}$,\n" + \
                                  "baseline_gp_matern32_lnrho_flux_lc," + str(distribution_params.lnrho) + ",1," + \
                                  distribution + " " + str(distribution_params.lnrho_lower_err) + " " + \
                                  str(distribution_params.lnrho_upper_err) + "," \
                                  "$\\\\mathrm{gp: \\\\ln{\\\\rho} (lc)}$,"
            else:
                sigma_lnyerr = (distribution_params.yerr_lower_err \
                    if distribution_params.yerr_lower_err > distribution_params.yerr_upper_err \
                    else distribution_params.yerr_upper_err) * self.yerr_err_multi
                sigma_lnsigma = (distribution_params.lnsigma_lower_err \
                    if distribution_params.lnsigma_lower_err > distribution_params.lnsigma_upper_err \
                    else distribution_params.lnsigma_upper_err) * self.sigma_err_multi
                sigma_lnrho = (distribution_params.lnrho_lower_err \
                    if distribution_params.lnrho_lower_err > distribution_params.lnrho_upper_err \
                    else distribution_params.lnrho_upper_err) * self.rho_err_multi
                errs_params = "ln_err_flux_lc," + str(distribution_params.yerr) + ",1," + distribution + " " + \
                              str(distribution_params.yerr) + " " + \
                              str(sigma_lnyerr) + "," \
                              "$\\\\log{\\\\sigma_\\\\mathrm{lc}}$,\\$\\\\log{ \\\\mathrm{rel. flux.} }$"
                baseline_params = "baseline_gp_matern32_lnsigma_flux_lc," + str(distribution_params.lnsigma) + ",1," + distribution + \
                                  " " + str(distribution_params.lnsigma) + " " + str(sigma_lnsigma) + "," \
                                  "$\\\\mathrm{gp: \\\\ln{\\\\sigma} (lc)}$,\n" + \
                                  "baseline_gp_matern32_lnrho_flux_lc," + str(distribution_params.lnrho) + ",1," + distribution + \
                                  " " + str(distribution_params.lnrho) + " " + str(sigma_lnrho) + "," \
                                  "$\\\\mathrm{gp: \\\\ln{\\\\rho} (lc)}$,"
        else:
            baseline_params = ""
        instrument_params = re.sub('\\${sherlock:err_params}', errs_params, instrument_params)
        instrument_params = re.sub('\\${sherlock:baseline_params}', baseline_params, instrument_params)
        return instrument_params



import numpy as np

from foldedleastsquares.stats import spectra
from sherlockpipe.scoring.SignalSelector import SignalSelector
from sherlockpipe.scoring.helper import compute_border_score, harmonic_spectrum
from sherlockpipe.search.transitresult import TransitResult
import foldedleastsquares as tls


class AverageSpectrumSignalSelector(SignalSelector):
    """
    Selects the signal with best SNR by summing all the residuals from each detrend search and then computing the SDE.
    """
    def __init__(self):
        super().__init__()

    def select(self, id_run, sherlock_target, star_info, transits_min_count, time, lcs, transit_results, wl, cadence):
        non_nan_result_args = np.argwhere(np.array([1 if result.results is not None else 0 for result in transit_results.values()]) == 1).flatten()
        non_nan_results_count = len(non_nan_result_args)
        if non_nan_results_count == 0:
            return AvgSpectrumSignalSelection(0, 0, TransitResult(None, None, 0, 0, 0, 0, [], [], 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, [], False))
        chi2_sum = np.nansum([result.results['chi2'] if result.results is not None else np.zeros(len(transit_results[non_nan_result_args[0]].results.chi2)) for result in transit_results.values()], axis=0) / non_nan_results_count
        SR, power_raw, power, SDE_raw, SDE = spectra(chi2_sum, sherlock_target.oversampling, len(time))
        index_highest_power = np.nanargmax(power)
        signals_powers_for_period = [result.results['power'][index_highest_power] if result.results is not None else np.nan for result in
                                     transit_results.values()]
        best_curve_for_signal = np.nanargmax(signals_powers_for_period)
        period = transit_results[non_nan_result_args[0]].results['periods'][index_highest_power]
        oversampling = sherlock_target.oversampling
        model = tls.transitleastsquares(time, lcs[best_curve_for_signal])
        power_args = {"transit_template": sherlock_target.fit_method, "n_transits_min": transits_min_count,
                      "T0_fit_margin": sherlock_target.t0_fit_margin, "show_progress_bar": False,
                      "use_threads": 1, "oversampling_factor": oversampling,
                      "duration_grid_step": sherlock_target.duration_grid_step,
                      "period_grid": np.array([period])}
        if star_info.ld_coefficients is not None:
            power_args["u"] = star_info.ld_coefficients
        power_args["R_star"] = star_info.radius
        power_args["R_star_min"] = star_info.radius_min
        power_args["R_star_max"] = star_info.radius_max
        power_args["M_star"] = star_info.mass
        power_args["M_star_min"] = star_info.mass_min
        power_args["M_star_max"] = star_info.mass_max
        if sherlock_target.custom_transit_template is not None:
            power_args["transit_template_generator"] = sherlock_target.custom_transit_template
        results = model.power(**power_args)
        depths = (1 - results.transit_depths) * 1000
        depths_err = results.transit_depths_uncertainties * 1000
        t0s = np.array(results.transit_times)
        in_transit = tls.transit_mask(time, period, results.duration, results.T0)
        transit_count = results.distinct_transit_count
        border_score = compute_border_score(time, results, in_transit, cadence)
        results.periods = transit_results[best_curve_for_signal].results['periods']
        results.power = power
        harmonic_power = harmonic_spectrum(results.periods, power)
        depth_err = np.sqrt(np.nansum([depth_err ** 2 for depth_err in depths_err])) / len(depths_err)
        result = TransitResult(power_args, results, period, period / 100, results['duration'],
                               results['T0'], t0s, depths,
                               depths_err, results['depth'], depth_err, results['odd_even_mismatch'],
                               (1 - results.depth_mean_even[0]) * 1000, (1 - results.depth_mean_odd[0]) * 1000,
                               transit_count, results.snr, SDE, None, border_score, in_transit, False,
                               harmonic_power)
        if result.snr > sherlock_target.snr_min and result.sde > sherlock_target.sde_min:  # and SDE[a] > SDE_min and FAP[a] < FAP_max):
            best_signal_score = 1
        else:
            best_signal_score = 0
        return AvgSpectrumSignalSelection(best_signal_score, best_curve_for_signal, result)

class AvgSpectrumSignalSelection:
    def __init__(self, score, curve_index, transit_result):
        self.score = score
        self.curve_index = curve_index
        self.transit_result = transit_result

    def get_message(self):
        curve_name = "PDCSAP_FLUX" if self.curve_index == 0 else str(self.curve_index - 1)
        return "Chosen signal with AVERAGE-SPECTRUM algorithm --> NAME: " + curve_name + \
               "\tPeriod:" + str(self.transit_result.period) + \
               "\tSNR: " + str(self.transit_result.snr) + \
               "\tSDE: " + str(self.transit_result.sde) + \
               "\tFAP: " + str(self.transit_result.fap) + \
               "\tBORDER_SCORE: " + str(self.transit_result.border_score)



import logging
import os
import sys

import json
import pandas as pd
from numpy import arange

from sherlockpipe.loading import common
from sherlockpipe.loading.common import get_from_dict
from sherlockpipe.system_stability.megno import MegnoStabilityCalculator
from sherlockpipe.system_stability.stability_calculator import PlanetInput


def run_stability(args):
    object_dir = os.getcwd() if args.object_dir is None else args.object_dir
    index = 0
    stability_dir = object_dir + "/stability_" + str(index)
    while os.path.exists(stability_dir) or os.path.isdir(stability_dir):
        stability_dir = object_dir + "/stability_" + str(index)
        index = index + 1
    os.mkdir(stability_dir)
    file_dir = stability_dir + "/stability.log"
    if os.path.exists(file_dir):
        os.remove(file_dir)
    if not isinstance(logging.root, logging.RootLogger):
        logging.root = logging.RootLogger(logging.INFO)
    formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger()
    while len(logger.handlers) > 0:
        logger.handlers.pop()
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    handler = logging.FileHandler(file_dir)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logging.info("Starting stability validation")
    star_df = pd.read_csv(object_dir + "/params_star.csv")
    star_mass = star_df.iloc[0]["M_star"]
    star_mass_low_err = star_df.iloc[0]["M_star_lerr"]
    star_mass_up_err = star_df.iloc[0]["M_star_uerr"]
    star_mass_low = star_mass if star_mass_low_err is None else star_mass - star_mass_low_err
    star_mass_up = star_mass if star_mass_up_err is None else star_mass + star_mass_up_err
    star_mass_bins = args.star_mass_bins
    free_params = args.free_params.split(",") if args.free_params is not None else []
    planets_params = []
    if args.properties is None:
        candidates = pd.read_csv(object_dir + "/../candidates.csv")
        ns_derived_file = object_dir + "/results/ns_derived_table.csv"
        ns_file = object_dir + "/results/ns_table.csv"
        if not os.path.exists(ns_derived_file) or not os.path.exists(ns_file):
            raise ValueError("Bayesian fit posteriors files {" + ns_file + ", " + ns_derived_file + "} not found")
        fit_derived_results = pd.read_csv(object_dir + "/results/ns_derived_table.csv")
        fit_results = pd.read_csv(object_dir + "/results/ns_table.csv")
        candidates_count = len(fit_results[fit_results["#name"].str.contains("_period")])
        ecc_rows = fit_derived_results[fit_derived_results["#property"].str.contains("Eccentricity")]
        arg_periastron_rows = fit_derived_results[
            fit_derived_results["#property"].str.contains("Argument of periastron")]
        for i in arange(0, candidates_count):
            period_row = fit_results[fit_results["#name"].str.contains("_period")].iloc[i]
            period = float(period_row["median"])
            period_low_err = float(period_row["lower_error"])
            period_up_err = float(period_row["upper_error"])
            inc_row = fit_derived_results[fit_derived_results["#property"].str.contains("Inclination")].iloc[i]
            inclination = float(inc_row["value"])
            inc_low_err = float(inc_row["lower_error"])
            inc_up_err = float(inc_row["upper_error"])
            duration_row = \
            fit_derived_results[fit_derived_results["#property"].str.contains("Total transit duration")].iloc[i]
            duration = float(duration_row["value"])
            duration_low_err = float(duration_row["lower_error"])
            duration_up_err = float(duration_row["upper_error"])
            radius_row = fit_derived_results[fit_derived_results["#property"].str.contains("R_\{\\\\oplus}")].iloc[i]
            radius = float(radius_row["value"])
            radius_low_err = float(radius_row["lower_error"])
            radius_up_err = float(radius_row["upper_error"])
            if len(ecc_rows) > 0:
                ecc_row = ecc_rows.iloc[i]
                eccentricity = float(ecc_row["value"])
                ecc_low_err = float(ecc_row["lower_error"])
                ecc_up_err = float(ecc_row["upper_error"])
            else:
                eccentricity = 0.0
                ecc_low_err = 0.1
                ecc_up_err = 0.1
            if len(arg_periastron_rows) > 0:
                arg_periastron_row = arg_periastron_rows.iloc[i]
                arg_periastron = float(arg_periastron_row["value"])
                arg_periastron_low_err = float(arg_periastron_row["lower_error"])
                arg_periastron_up_err = float(arg_periastron_row["upper_error"])
            else:
                arg_periastron = 0.0
                arg_periastron_low_err = 20.0
                arg_periastron_up_err = 20.0
            planets_params.append(
                PlanetInput(period=period, period_low_err=period_low_err, period_up_err=period_up_err,
                            radius=radius, radius_low_err=radius_low_err, radius_up_err=radius_up_err,
                            eccentricity=eccentricity, ecc_low_err=ecc_low_err, ecc_up_err=ecc_up_err,
                            inclination=inclination, inc_low_err=inc_low_err, inc_up_err=inc_up_err,
                            omega=arg_periastron, omega_low_err=arg_periastron_low_err,
                            omega_up_err=arg_periastron_up_err, mass_bins=args.mass_bins, period_bins=args.period_bins,
                            ecc_bins=args.ecc_bins, inc_bins=args.inc_bins, omega_bins=args.omega_bins))
    else:
        user_properties = common.load_from_yaml(args.properties)
        user_planet_params = []
        for planet in user_properties["BODIES"]:
            period_bins = args.period_bins if "P_BINS" not in planet else planet["P_BINS"]
            ecc_bins = args.ecc_bins if "E_BINS" not in planet else planet["E_BINS"]
            mass_bins = args.mass_bins if "M_BINS" not in planet else planet["M_BINS"]
            inc_bins = args.inc_bins if "I_BINS" not in planet else planet["I_BINS"]
            om_bins = args.omega_bins if "O_BINS" not in planet else planet["O_BINS"]
            user_planet_params.append(PlanetInput(period=get_from_dict(planet, "P"),
                                                  period_low_err=get_from_dict(planet, "P_LOW"),
                                                  period_up_err=get_from_dict(planet, "P_UP"),
                                                  radius=get_from_dict(planet, "R"),
                                                  radius_low_err=get_from_dict(planet, "R_LOW"),
                                                  radius_up_err=get_from_dict(planet, "R_UP"),
                                                  mass=get_from_dict(planet, "M"),
                                                  mass_low_err=get_from_dict(planet, "M_LOW"),
                                                  mass_up_err=get_from_dict(planet, "M_UP"),
                                                  eccentricity=get_from_dict(planet, "E_LOW"),
                                                  ecc_low_err=get_from_dict(planet, "E_LOW"),
                                                  ecc_up_err=get_from_dict(planet, "E_UP"),
                                                  inclination=get_from_dict(planet, "I"),
                                                  inc_low_err=get_from_dict(planet, "I_LOW"),
                                                  inc_up_err=get_from_dict(planet, "I_UP"),
                                                  omega=get_from_dict(planet, "O"),
                                                  omega_low_err=get_from_dict(planet, "O_LOW"),
                                                  omega_up_err=get_from_dict(planet, "O_UP"),
                                                  period_bins=period_bins, mass_bins=mass_bins, ecc_bins=ecc_bins,
                                                  inc_bins=inc_bins, omega_bins=om_bins))
        planets_params = planets_params + user_planet_params
        if "STAR" in user_properties:
            star_mass = star_mass_low if "M" not in user_properties["STAR"] else user_properties["STAR"]["M"]
            star_mass_low = star_mass_low if "M_LOW" not in user_properties["STAR"] else star_mass - \
                                                                                         user_properties["STAR"][
                                                                                             "M_LOW"]
            star_mass_up = star_mass_up if "M_UP" not in user_properties["STAR"] else star_mass + \
                                                                                      user_properties["STAR"]["M_UP"]
            star_mass_bins = args.star_mass_bins if "M_BINS" not in user_properties["STAR"] else \
            user_properties["STAR"]["M_BINS"]
    stability_calculator = MegnoStabilityCalculator(args.years)
    logger.info("%.0f planets to be simulated", len(planets_params))
    logger.info("Lowest star mass: %.2f", star_mass_low)
    logger.info("Highest star mass: %.2f", star_mass_up)
    logger.info("Star mass bins: %.0f", star_mass_bins)
    for key, body in enumerate(planets_params):
        logger.info("Body %.0f: %s", key, json.dumps(body.__dict__))
    stability_calculator.run(stability_dir, star_mass_low, star_mass_up, star_mass_bins, planets_params, args.cpus, free_params)


import random
from sherlockpipe.scoring.SignalSelector import SignalSelector, SignalSelection


class RandomSignalSelector(SignalSelector):
    """
    Implements a random signal selection from a list of detrends. Only for testing.
    """
    def __init__(self):
        super().__init__()

    def select(self, id_run, sherlock_target, star_info, transits_min_count, time, lcs, transit_results, wl, cadence):
        best_signal_snr_index = random.randrange(0, len(transit_results) - 1, 1)
        best_signal = transit_results[best_signal_snr_index]
        best_signal_snr = best_signal.snr
        if best_signal_snr > sherlock_target.snr_min:  # and SDE[a] > SDE_min and FAP[a] < FAP_max):
            best_signal_score = 1
        else:
            best_signal_score = 0
        # You could also extend SignalSelector class to provide more information about your choice.
        return SignalSelection(best_signal_score, best_signal_snr_index, best_signal)


import numpy as np

from sherlockpipe.scoring.BasicSignalSelector import BasicSignalSelector
from sherlockpipe.scoring.SignalSelector import SignalSelection


class SdeBorderCorrectedSignalSelector(BasicSignalSelector):
    """
    Selects the best signal among the SDE border-corrected signals. The applied correction consists in calculating how
    many transits occur in times very close to observation starts/ends.
    """
    def __init__(self):
        super().__init__()

    def select(self, id_run, sherlock_target, star_info, transits_min_count, time, lcs, transit_results, wl, cadence):
        basic_signal_selection = super().select(id_run, sherlock_target, star_info, transits_min_count, time, lcs,
                                                transit_results, wl, cadence)
        signals_sde = np.nan_to_num([transit_result.sde * transit_result.border_score
                                     for key, transit_result in transit_results.items()])
        best_signal_sde = np.nanmax(signals_sde)
        best_signal_sde_index = np.nanargmax(signals_sde)
        selected_signal_snr = transit_results[best_signal_sde_index].snr
        selected_signal = transit_results[best_signal_sde_index]
        if best_signal_sde > sherlock_target.sde_min and selected_signal_snr > sherlock_target.snr_min:  # and SDE[a] > SDE_min and FAP[a] < FAP_max):
            best_signal_score = 1
        else:
            best_signal_score = 0
        return CorrectedBorderSdeSignalSelection(best_signal_score, best_signal_sde, basic_signal_selection.curve_index,
                                              transit_results[basic_signal_selection.curve_index],
                                              best_signal_sde_index, selected_signal)


class CorrectedBorderSdeSignalSelection(SignalSelection):
    def __init__(self, score, corrected_sde, original_curve_index, original_transit_result, final_curve_index,
                 final_transit_result):
        super().__init__(score, final_curve_index, final_transit_result)
        self.original_curve_index = original_curve_index
        self.original_transit_result = original_transit_result
        self.corrected_sde = corrected_sde

    def get_message(self):
        curve_name = "PDCSAP_FLUX" if self.curve_index == 0 else str(self.curve_index - 1)
        original_curve_name = "PDCSAP_FLUX" if self.original_curve_index == 0 else str(self.original_curve_index - 1)
        return "Chosen signal with BORDER_CORRECT algorithm --> NAME: " + curve_name + \
               "\tPeriod:" + str(self.transit_result.period) + \
               "\tCORR_SDE: " + str(self.corrected_sde) + \
               "\tSNR: " + str(self.transit_result.snr) + \
               "\tSDE: " + str(self.transit_result.sde) + \
               "\tFAP: " + str(self.transit_result.fap) + \
               "\tBORDER_SCORE: " + str(self.transit_result.border_score) + \
               "\nProposed selection with BASIC algorithm was --> NAME: " + original_curve_name + \
               "\tPeriod:" + str(self.original_transit_result.period) + \
               "\tSNR: " + str(self.original_transit_result.snr)



import logging
import math
import os
import sys
from argparse import ArgumentParser
import pandas as pd

from sherlockpipe.loading import common


def tidal_args_parse(args=None):
    ap = ArgumentParser(description='Validation of system stability')
    ap.add_argument('--object_dir', help="If the object directory is not your current one you need to provide the "
                                         "ABSOLUTE path", required=False)
    ap.add_argument('--candidate', type=int, default=None, help="The candidate signal to be used.", required=False)
    ap.add_argument('--properties', help="The YAML file to be used as input.", required=False)
    return ap.parse_args(args)


def run_tidal(object_dir, candidate, properties):
    object_dir = os.getcwd() if object_dir is None else object_dir
    candidates = pd.read_csv(object_dir + "/candidates.csv")
    if not isinstance(logging.root, logging.RootLogger):
        logging.root = logging.RootLogger(logging.INFO)
    formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger()
    while len(logger.handlers) > 0:
        logger.handlers.pop()
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logging.info("Starting tidal locking time computation")
    star_df = pd.read_csv(object_dir + "/params_star.csv")
    if candidate is None:
        user_properties = common.load_from_yaml(properties)
        candidate = pd.DataFrame(columns=['id', 'period', 'depth', 't0', 'sectors', 'number', 'lc'])
        candidate = candidate.append(user_properties, ignore_index=True)
        candidate['id'] = star_df.iloc[0]["obj_id"]
    else:
        candidate_selection = int(candidate)
        if candidate_selection < 1 or candidate_selection > len(candidates.index):
            raise SystemExit("User selected a wrong candidate number.")
        candidates = candidates.rename(columns={'Object Id': 'id'})
        candidate = candidates.iloc[[candidate_selection - 1]]
        candidate['number'] = [candidate_selection]
        logging.info("Selected signal number " + str(candidate_selection))
    star_df = pd.read_csv(object_dir + "/params_star.csv")
    star_mass = star_df.iloc[0]["M_star"]
    star_mass_low_err = star_df.iloc[0]["M_star_lerr"]
    star_mass_up_err = star_df.iloc[0]["M_star_uerr"]
    logger.info("Star mass: %.2f", star_mass)
    logger.info("Lowest star mass: %.2f", star_mass - star_mass_low_err)
    logger.info("Highest star mass: %.2f", star_mass + star_mass_up_err)
    G = 6.67259e-11  # gravitational constant
    Q = 100  # planet's tidal dissipation function
    semimajor_axis = candidate.iloc[0]['a']
    rp = candidate.iloc[0]["rad_p"]
    planet_mean_motion = math.sqrt((G * (star_mass + 1)) / (semimajor_axis ** 3)) #angular velocity of the planet in radians per second
    T_rot_planet = (2 * math.pi) / planet_mean_motion
    # "Tidal Evolution of Close-in Planets" by Matthew J. Holman and Scott J. Tremaine, The Astrophysical Journal, vol. 552, pp. 693–718 (2001)
    # This paper provides a detailed derivation of the tidal evolution equations, including the time scale for a planet to become tidally locked.
    T_lock = (3 * math.sqrt(T_rot_planet) * semimajor_axis ** 6) / (Q * G * star_mass * rp ** 5)
    print("Tidal Locking Time: {:.2e} years".format(T_lock))


if __name__ == '__main__':
    args = tidal_args_parse()
    run_tidal(args.object_dir, args.candidate, args.properties)



import logging

import foldedleastsquares
import numpy as np
from lcbuilder.helper import LcbuilderHelper


class ToolWithCandidate:
    """
    Base class to be extended by tools that process candidates information.
    """

    def __init__(self, is_candidate_from_search, candidates_df) -> None:
        super().__init__()
        self.is_candidate_from_search = is_candidate_from_search
        self.search_candidates_df = candidates_df

    def mask_previous_candidates(self, time, flux, flux_err, candidate_id):
        """
        Masks all the candidates found in previous runs in the SHERLOCK search.

        :param time: the time array
        :param flux: the flux measurements array
        :param flux_err: the flux error measurements array
        :param candidate_id: the candidate number
        :return: time, flux and flux_err with previous candidates in-transit data masked
        """
        if self.is_candidate_aware():
            for index in np.arange(0, candidate_id - 1):
                candidate_row = self.search_candidates_df.iloc[index]
                period = candidate_row["period"]
                duration = candidate_row["duration"]
                duration = duration / 60 / 24
                t0 = candidate_row["t0"]
                logging.info("Masking candidate number %.0f with P=%.3fd, T0=%.2f and D=%.2fd", index + 1, period, t0,
                             duration)
                time, flux, flux_err = LcbuilderHelper.mask_transits(time, flux, period, duration * 2, t0, flux_err)
        return time, flux, flux_err

    def is_candidate_aware(self):
        """
        Boolean return to inform whether the candidate to be processed comes from SHERLOCK searches or is user-given.

        :return: boolean with the value
        """
        return self.is_candidate_from_search



"""Run script for observation plan"""

import shutil
import os

import alexfitter
import numpy as np
import pandas as pd
from lcbuilder.lcbuilder_class import LcBuilder
from sherlockpipe.observation_plan.observation_report import ObservationReport
from sherlockpipe.observation_plan.planner import Planner
from astropy.time import Time


def run_plan(args):
    if args.observatories is None and (args.lat is None or args.lon is None or args.alt is None):
        raise ValueError("You either need to set the 'observatories' property or the lat, lon and alt.")
    object_dir = os.getcwd() if args.object_dir is None else args.object_dir
    ns_derived_file = object_dir + "/results/ns_derived_table.csv"
    ns_file = object_dir + "/results/ns_table.csv"
    if not os.path.exists(ns_derived_file) or not os.path.exists(ns_file):
        raise ValueError("Bayesian fit posteriors files {" + ns_file + ", " + ns_derived_file + "} not found")
    star_df = pd.read_csv(object_dir + "/params_star.csv")
    object_id = star_df.iloc[0]["obj_id"]
    ra = star_df.iloc[0]["ra"]
    dec = star_df.iloc[0]["dec"]
    fit_derived_results = pd.read_csv(object_dir + "/results/ns_derived_table.csv")
    fit_results = pd.read_csv(object_dir + "/results/ns_table.csv")
    candidates_count = len(fit_results[fit_results["#name"].str.contains("_period")])
    alles = alexfitter.allesclass(object_dir)
    since = Time.now() if args.since is None else Time(args.since, scale='utc')
    percentile = 99.7 if args.error_sigma == 3 else 95 if args.error_sigma == 2 else 68
    for i in np.arange(0, candidates_count):
        period_row = fit_results[fit_results["#name"].str.contains("_period")].iloc[i]
        period = period_row["median"]
        period_distribution = alles.posterior_params[period_row["#name"]]
        period_low_err = period - np.percentile(period_distribution, 50 - percentile / 2)
        period_up_err = np.percentile(period_distribution, 50 + percentile / 2) - period
        name = object_id + "_" + period_row["#name"].replace("_period", "")
        epoch_row = fit_results[fit_results["#name"].str.contains("_epoch")].iloc[i]
        epoch = epoch_row["median"].item()
        epoch_distribution = alles.posterior_params[epoch_row["#name"]]
        epoch_low_err = epoch - np.percentile(epoch_distribution, 50 - percentile / 2)
        epoch_up_err = np.percentile(epoch_distribution, 50 + percentile / 2) - epoch
        duration_row = fit_derived_results[fit_derived_results["#property"].str.contains("Total transit duration")].iloc[i]
        duration = duration_row["value"].item()
        duration_low_err = float(duration_row["lower_error"])
        duration_up_err = float(duration_row["upper_error"])
        depth_row = fit_derived_results[fit_derived_results["#property"].str.contains("depth \(dil.\)")].iloc[i]
        depth = depth_row["value"] * 1000
        depth_low_err = depth_row["lower_error"] * 1000
        depth_up_err = depth_row["upper_error"] * 1000
        mission, mission_prefix, id_int = LcBuilder().parse_object_info(object_id)
        if args.time_unit is None and mission == "TESS":
            time_unit = 'btjd'
        elif args.time_unit is None and mission == "Kepler" or mission == "K2":
            time_unit = 'bkjd'
        elif args.time_unit is not None:
            time_unit = args.time_unit
        else:
            time_unit = 'jd'
        observatories_df, observables_df, alert_date, plan_dir, images_dir = \
            Planner.create_observation_observables(object_id, object_dir, ra, dec, since, name, epoch,
                                                   epoch_low_err, epoch_up_err, period, period_low_err,
                                                   period_up_err, duration, args.observatories, args.tz, args.lat,
                                                   args.lon, args.alt, args.max_days, args.min_altitude,
                                                   args.moon_min_dist, args.moon_max_dist, args.transit_fraction,
                                                   args.baseline, not args.no_error_alert, time_unit)
        report = ObservationReport(observatories_df, observables_df, alert_date, object_id, name, plan_dir, ra, dec,
                                   epoch, epoch_low_err, epoch_up_err, period, period_low_err, period_up_err, duration,
                                   duration_low_err, duration_up_err, depth, depth_low_err, depth_up_err,
                                   args.transit_fraction, args.moon_min_dist, args.moon_max_dist, args.min_altitude,
                                   args.max_days, star_df.iloc[0]["v"], star_df.iloc[0]["j"], star_df.iloc[0]["h"],
                                   star_df.iloc[0]["k"])
        report.create_report()
        shutil.rmtree(images_dir, ignore_errors=True)
